#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4

import os
import shutil
import glob


def clean():
    to_delete = [
        'session_stats',
        'libtorrent_logs*',
        'round_trip_ms.log',
        'dht.log',
        'upnp.log',
        'natpmp.log',
        'bin',
        'build-aux',
        '.deps',
        'test_tmp_*',
        'bjam_build.*.xml',
        '*.exe',
        '*.pdb',
        '*.pyd',
        'dist',
        'build',
        '.libs',
        '*.cpp.orig',
        '*.cpp.rej',
        '*.hpp.orig',
        '*.hpp.rej',
        '*.gcov',
        '*.gcno',
        '*.gcda',
        'lib*.a',
        'Jamfile.rej',
        'Jamfile.orig',
        '*.o',
        '*.lo',
        'autom4te.cache',
        'configure',
        'config.report',
        'config.log',
        '.lib',
        'CMakeFiles',
        'CMakeCache.txt',
        'checking_benchmark',
        'cpu_benchmark',
    ]

    directories = [
        'examples',
        'test',
        '.',
        'tools',
        'src',
        'simulation',
        'fuzzers',
        os.path.join('src', 'kademlia'),
        os.path.join('include', 'libtorrent'),
        os.path.join('include', os.path.join('libtorrent', '_aux')),
        os.path.join('include', os.path.join('libtorrent', 'kademlia')),
        os.path.join('bindings', 'python'),
        os.path.join('bindings', os.path.join('python', 'src')),
        os.path.join('bindings', 'c'),
        os.path.join('bindings', os.path.join('c', 'src')),
        os.path.join('simulation', 'libsimulator')
    ]

    for d in directories:
        for f in to_delete:
            path = os.path.join(d, f)
            entries = glob.glob(path)
            for p in entries:
                try:
                    shutil.rmtree(p)
                    print(p)
                except Exception as e:
                    print(p, e)
                    try:
                        os.remove(p)
                        print(p)
                    except Exception as e:
                        print(p, e)


if __name__ == "__main__":
    clean()



#!/usr/bin/python

# essentially copy pasted from http://0pointer.de/blog/projects/copyright.html

from subprocess import Popen, PIPE
from datetime import datetime


def pretty_years(s):

    li = list(s)
    li.sort()

    start = None
    prev = None
    r = []

    for x in li:
        if prev is None:
            start = x
            prev = x
            continue

        if x == prev + 1:
            prev = x
            continue

        if prev == start:
            r.append("%i" % prev)
        else:
            r.append("%i-%i" % (start, prev))

        start = x
        prev = x

    if prev is not None:
        if prev == start:
            r.append("%i" % prev)
        else:
            r.append("%i-%i" % (start, prev))

    return ", ".join(r)


def order_by_year(a):

    la = list(a[2])
    la.sort()
    # sort by year first, then by author
    return (la[0], a[0])


author_map = {
    'arvidn': 'Arvid Norberg',
    'pavel.pimenov': 'Pavel Pimenov',
    'd_komarov': 'd-komarov',
    'Chocobo1': 'Mike Tzou',
    'unsh': 'Un Shyam',
    'toinetoine': 'Antoine Dahan'
}


def map_author(a):
    if a in author_map:
        return author_map[a]
    else:
        return a


def get_authors(f):

    print("File: %s" % f)

    commits = []
    data = {}

    for ln in Popen(["git", "blame", "--incremental", f], stdout=PIPE).stdout:

        ln = ln.decode('utf8')
        if ln.startswith("filename "):
            if len(data) > 0:
                commits.append(data)
            data = {}

        elif ln.startswith("author "):
            data["author"] = map_author(ln[7:].strip())

        elif ln.startswith("author-mail <"):
            data["author-mail"] = ln[12:].strip()

        elif ln.startswith("author-time "):
            data["author-time"] = ln[11:].strip()

        elif ln.startswith("author-tz "):
            data["author-tz"] = ln[9:].strip()

    by_author = {}

    for c in commits:
        try:
            if c['author'] == 'Not Committed Yet':
                continue
            n = by_author[c["author"]]
        except KeyError:
            n = (c["author"], c["author-mail"], set())
            by_author[c["author"]] = n

        # FIXME: Handle time zones properly
        year = datetime.fromtimestamp(int(c["author-time"])).year

        n[2].add(year)

    for an, a in list(by_author.items()):
        for bn, b in list(by_author.items()):
            if a is b:
                continue

            if a[1] == b[1]:
                a[2].update(b[2])

                if an in by_author and bn in by_author:
                    del by_author[bn]

    copyright = list(by_author.values())
    copyright.sort(key=order_by_year)

    ret = ''
    for name, mail, years in copyright:
        ret += "Copyright (c) %s, %s\n" % (pretty_years(years), name)
    return ret



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4

import argparse
import os
import platform
import shutil
import subprocess
import sys
import time

from vmstat import capture_sample
from vmstat import plot_output
from vmstat import print_output_to_file


def main():
    args = parse_args()

    ret = os.system(f"cd ../examples && b2 release {args.toolset} stage_client_test stage_connection_tester")
    save_dir = args.directory
    print(f"save dir: {save_dir}")

    if ret != 0:
        print('ERROR: build failed: %d' % ret)
        sys.exit(1)

    rm_file_or_dir(".ses_state")
    rm_file_or_dir(".resume")

    if not os.path.exists('checking_benchmark.torrent'):
        ret = os.system('../examples/connection_tester gen-torrent -s 10000 -n 15 -t checking_benchmark.torrent')
        if ret != 0:
            print('ERROR: connection_tester failed: %d' % ret)
            sys.exit(1)

    if not os.path.exists(f"{save_dir}/checking_benchmark"):
        cmd_line = f'../examples/connection_tester gen-data -t checking_benchmark.torrent -P {save_dir}'
        print(cmd_line)
        ret = os.system(cmd_line)
        if ret != 0:
            print('ERROR: connection_tester failed: %d' % ret)
            sys.exit(1)

    for threads in [1, 2, 4, 8, 16, 32, 64]:
        print("drop caches now. e.g. \"echo 1 | sudo tee /proc/sys/vm/drop_caches\"")
        input("Press Enter to continue...")
        run_test(f"{threads}", f"--hashing_threads={threads}", save_dir)


def run_test(name, client_arg, save_dir: str):
    output_dir = 'logs_checking_%s' % name

    timing_path = os.path.join(output_dir, 'timing.txt')
    if os.path.exists(timing_path):
        print('file "{path}" exists, skipping test "{name}"'.format(path=timing_path, name=name))
        return

    rm_file_or_dir(output_dir)
    try:
        os.mkdir(output_dir)
    except Exception:
        pass

    rm_file_or_dir(f"{save_dir}/.resume")

    client_cmd = ('../examples/client_test checking_benchmark.torrent '
        '--enable_dht=0 --enable_lsd=0 --enable_upnp=0 --enable_natpmp=0 '
        f'-1 {client_arg} -s {save_dir} -f {output_dir}/events.log --alert_mask=all')

    client_out = open('%s/client.out' % output_dir, 'w+')
    print('client_cmd: "{cmd}"'.format(cmd=client_cmd))
    c = subprocess.Popen(client_cmd.split(' '), stdout=client_out, stderr=client_out, stdin=subprocess.PIPE)

    start = time.monotonic()
    if platform.system() == "Linux":
        out = {}
        while c.returncode is None:
            capture_sample(c.pid, start, out)
            time.sleep(0.1)
            c.poll()

        stats_filename = f"{output_dir}/memory_stats.log"
        keys = print_output_to_file(out, stats_filename)
        plot_output(stats_filename, keys)
    else:
        c.wait()

    client_out.close()

    start_time = 0
    end_time = 0
    for l in open('%s/events.log' % output_dir, 'r'):
        if 'checking_benchmark: start_checking, m_checking_piece: ' in l \
                and start_time == 0:
            start_time = int(l.split(' ')[0][1:-1])
        if 'state changed to: finished' in l \
                and start_time != 0:
            end_time = int(l.split(' ')[0][1:-1])

    print('%s: %d' % (name, end_time - start_time))
    with open('%s/timing.txt' % output_dir, 'w+') as f:
        f.write('%s: %d\n' % (name, end_time - start_time))


def rm_file_or_dir(path):
    """ Attempt to remove file or directory at path
    """
    try:
        shutil.rmtree(path)
    except Exception:
        pass

    try:
        os.remove(path)
    except Exception:
        pass


def parse_args():
    p = argparse.ArgumentParser()
    p.add_argument('--toolset', default="")
    p.add_argument('--directory', default=".")

    return p.parse_args()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
from __future__ import print_function

counter_types = {}

f = open('../include/libtorrent/performance_counters.hpp')

counter_type = ''

for line in f:

    # ignore anything after //
    if '//' in line:
        line = line.split('//')[0]

    line = line.strip()

    if line.startswith('#'):
        continue
    if line == '':
        continue

    if 'enum stats_counter_t' in line:
        counter_type = 'counter'
        continue

    if 'enum stats_gauge_t' in line:
        counter_type = 'gauge'
        continue

    if '{' in line or '}' in line or 'struct' in line or 'namespace' in line:
        continue
    if counter_type == '':
        continue
    if not line.endswith(','):
        continue

    # strip off trailing comma
    line = line[:-1]
    if '=' in line:
        line = line[:line.index('=')].strip()

    counter_types[line] = counter_type

f.close()

f = open('../src/session_stats.cpp')

out = open('stats_counters.rst', 'w+')


def print_field(str, width):
    return '%s%s' % (str, ' ' * (width - len(str)))


def render_section(names, description, types):
    max_name_len = max(len(max(names, key=len)), len('name'))
    max_type_len = max(len(max(types, key=len)), len('type'))

    if description == '':
        for n in names:
            print('WARNING: no description for "%s"' % n)

    # add link targets for the rest of the manual to reference
    for n in names:
        print('.. _%s:\n' % n, file=out)

    if len(names) > 0:
        print('.. raw:: html\n', file=out)
        for n in names:
            print('\t<a name="%s"></a>' % n, file=out)
        print('', file=out)

    separator = '+-' + ('-' * max_name_len) + '-+-' + ('-' * max_type_len) + '-+'

    # build a table for the settings, their type and default value
    print(separator, file=out)
    print('| %s | %s |' % (print_field('name', max_name_len), print_field('type', max_type_len)), file=out)
    print(separator.replace('-', '='), file=out)
    for i in range(len(names)):
        print('| %s | %s |' % (print_field(names[i], max_name_len), print_field(types[i], max_type_len)), file=out)
        print(separator, file=out)
    print(file=out)
    print(description, file=out)
    print('', file=out)


mode = ''

description = ''
names = []
types = []

for line in f:
    description_line = line.lstrip().startswith('//')

    line = line.strip()

    if mode == 'ignore':
        if '#endif' in line:
            mode = ''
        continue

    if 'TORRENT_ABI_VERSION == 1' in line:
        mode = 'ignore'
        continue

    if description_line:
        if len(names) > 0:
            render_section(names, description, types)
            description = ''
            names = []
            types = []

        description += '\n' + line[3:]

    if '#define' in line:
        continue

    if 'METRIC(' in line:
        args = line.split('(')[1].split(')')[0].split(',')

        # args: category, name, type

        args[1] = args[1].strip()
        names.append(args[0].strip() + '.' + args[1].strip())
        types.append(counter_types[args[1]])

if len(names) > 0:
    render_section(names, description, types)

out.close()
f.close()



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4

from __future__ import print_function

import os
import sys
from functools import reduce

# usage: parse_log.py log-file [socket-index to focus on]


socket_filter = None
if len(sys.argv) >= 3:
    socket_filter = sys.argv[2].strip()

if socket_filter is None:
    print("scanning for socket with the most packets")
    file = open(sys.argv[1], 'r')

    sockets = {}

    for line in file:
        if 'our_delay' not in line:
            continue

        try:
            a = line.strip().split(" ")
            socket_index = a[1][:-1]
        except Exception:
            continue

        # msvc's runtime library doesn't prefix pointers
        # with '0x'
        # if socket_index[:2] != '0x':
        #     continue

        if socket_index in sockets:
            sockets[socket_index] += 1
        else:
            sockets[socket_index] = 1

    items = list(sockets.items())
    items.sort(reverse=True, key=lambda x: x[1])

    count = 0
    for i in items:
        print('%s: %d' % (i[0], i[1]))
        count += 1
        if count > 5:
            break

    file.close()
    socket_filter = items[0][0]
    print('\nfocusing on socket %s' % socket_filter)

file = open(sys.argv[1], 'r')
out_file = 'utp.out%s' % socket_filter
out = open(out_file, 'w')

delay_samples = 'points lc rgb "blue"'
delay_base = 'steps lw 2 lc rgb "purple"'
target_delay = 'steps lw 2 lc rgb "red"'
off_target = 'dots lc rgb "blue"'
cwnd = 'steps lc rgb "green" lw 2'
window_size = 'steps lc rgb "sea-green"'
rtt = 'lines lc rgb "light-blue"'
send_buffer = 'lines lc rgb "light-red"'

metrics = {
    'our_delay': ['our delay (ms)', 'x1y2', delay_samples],
    'upload_rate': ['send rate (B/s)', 'x1y1', 'lines'],
    'max_window': ['cwnd (B)', 'x1y1', cwnd],
    'target_delay': ['target delay (ms)', 'x1y2', target_delay],
    'cur_window': ['bytes in-flight (B)', 'x1y1', window_size],
    'cur_window_packets': ['number of packets in-flight', 'x1y2', 'steps'],
    'packet_size': ['current packet size (B)', 'x1y2', 'steps'],
    'rtt': ['rtt (ms)', 'x1y2', rtt],
    'off_target': ['off-target (ms)', 'x1y2', off_target],
    'delay_sum': ['delay sum (ms)', 'x1y2', 'steps'],
    'their_delay': ['their delay (ms)', 'x1y2', delay_samples],
    'get_microseconds': ['clock (us)', 'x1y1', 'steps'],
    'wnduser': ['advertised window size (B)', 'x1y1', 'steps'],
    'ssthres': ['slow-start threshold (B)', 'x1y1', 'steps'],
    'timeout': ['until next timeout (ms)', 'x1y2', 'steps'],
    'rto': ['current timeout (ms)', 'x1y2', 'steps'],

    'delay_base': ['delay base (us)', 'x1y1', delay_base],
    'their_delay_base': ['their delay base (us)', 'x1y1', delay_base],
    'their_actual_delay': ['their actual delay (us)', 'x1y1', delay_samples],
    'actual_delay': ['actual_delay (us)', 'x1y1', delay_samples],
    'send_buffer': ['send buffer size (B)', 'x1y1', send_buffer],
    'recv_buffer': ['receive buffer size (B)', 'x1y1', 'lines'],
    'packet_loss': ['packet lost', 'x1y2', 'steps'],
    'packet_timeout': ['packet timed out', 'x1y2', 'steps'],
    'acked_bytes': ['Bytes ACKed by packet', 'x1y2', 'steps'],
    'bytes_sent': ['cumulative bytes sent', 'x1y2', 'steps'],
    'bytes_resent': ['cumulative bytes resent', 'x1y2', 'steps'],
    'written': ['reported written bytes', 'x1y2', 'steps'],
    'ack_nr': ['acked sequence number', 'x1y2', 'steps'],
    'seq_nr': ['sent sequence number', 'x1y2', 'steps'],
}

histogram_quantization = 1.0
socket_index = None

columns = []

begin = None

title = "-"
packet_loss = 0
packet_timeout = 0
num_acked = 0

delay_histogram = {}
packet_size_histogram = {}
window_size = {'0': 0, '1': 0}
bytes_sent = 0
bytes_resent = 0
written = 0
ack_nr = 0
seq_nr = 0

# [35301484] 0x00ec1190: actual_delay:1021583 our_delay:102 their_delay:-1021345 off_target:297 max_window:2687
# upload_rate:18942 delay_base:1021481154 delay_sum:-1021242 target_delay:400 acked_bytes:1441 cur_window:2882
# scaled_gain:2.432

counter = 0

print("reading log file")

for line in file:
    if "UTP_Connect" in line:
        title = line[:-2]
        if socket_filter is not None:
            title += ' socket: %s' % socket_filter
        else:
            title += ' sum of all sockets'
        continue

    try:
        a = line.strip().split(" ")
        t = a[0][1:-1]
        socket_index = a[1][:-1]
    except Exception:
        continue
#    if socket_index[:2] != '0x':
#        continue

    if socket_filter is not None and socket_index != socket_filter:
        continue

    counter += 1
    if (counter % 300 == 0):
        print("\r%d  " % counter, end=' ')

    if "lost." in line:
        packet_loss += 1
        continue

    if "lost (timeout)" in line:
        packet_timeout += 1
        continue

    if "acked packet " in line:
        num_acked += 1

    if "sending packet" in line:
        v = line.split('size:')[1].split(' ')[0]
        packet_size_histogram[v] = 1 + packet_size_histogram.get(v, 0)
        bytes_sent += int(v)

    if "re-sending packet" in line:
        v = line.split('size:')[1].split(' ')[0]
        bytes_resent += int(v)

    if 'calling write handler' in line:
        v = line.split('written:')[1].split(' ')[0]
        written += int(v)

    if "incoming packet" in line \
            and "ERROR" not in line \
            and "seq_nr:" in line \
            and "type:ST_SYN" not in line:
        if "ack_nr:" not in line:
            print(line)
        ack_nr = int(line.split('ack_nr:')[1].split(' ')[0])
        seq_nr = int(line.split('seq_nr:')[1].split(' ')[0])

    if "our_delay:" not in line:
        continue

# used for Logf timestamps
#    t, m = t.split(".")
#    t = time.strptime(t, "%H:%M:%S")
#    t = list(t)
#    t[0] += 107
#    t = tuple(t)
#    m = float(m)
#    m /= 1000.0
#    t = time.mktime(t) + m

# used for tick count timestamps
    t = int(t)

    if begin is None:
        begin = t
    t = t - begin
    # print time. Convert from milliseconds to seconds
    print('%f\t' % (float(t) / 1000.), end=' ', file=out)

    # if t > 200000:
    #    break

    fill_columns = not columns
    for i in a[2:]:
        try:
            n, v = i.split(':')
        except Exception:
            continue
        v = float(v)
        if n == "our_delay":
            bucket = int(v / histogram_quantization)
            delay_histogram[bucket] = 1 + delay_histogram.get(bucket, 0)
        if n not in metrics:
            continue
        if fill_columns:
            columns.append(n)
        if n == "max_window":
            window_size[socket_index] = v
            print('%f\t' % int(reduce(lambda a, b: a + b, list(window_size.values()))), end=' ', file=out)
        else:
            print('%f\t' % v, end=' ', file=out)

    if fill_columns:
        columns += ['packet_loss', 'packet_timeout', 'bytes_sent', 'ack_nr',
                    'seq_nr', 'bytes_resent', 'written']
    print(float(packet_loss), float(packet_timeout), float(bytes_sent),
          ack_nr, seq_nr, float(bytes_resent), written, file=out)
    packet_loss = 0
    packet_timeout = 0
    num_acked = 0
    written = 0

out.close()

out = open('%s.histogram' % out_file, 'w')
for d, f in delay_histogram.items():
    print(float(d * histogram_quantization) + histogram_quantization / 2.0, f, file=out)
out.close()

out = open('%s_packet_size.histogram' % out_file, 'w')
for d, f in packet_size_histogram.items():
    print(d, f, file=out)
out.close()

plot = [
    {
        'data': ['max_window', 'send_buffer', 'cur_window', 'rtt'],
        'title': 'send-packet-size',
        'y1': 'Bytes',
        'y2': 'Time (ms)'
    },
    {
        'data': ['max_window', 'send_buffer', 'cur_window', 'written'],
        'title': 'bytes-written',
        'y1': 'Bytes',
        'y2': 'Time (ms)'
    },
    {
        'data': ['upload_rate', 'max_window', 'cur_window', 'wnduser', 'cur_window_packets', 'packet_size', 'rtt'],
        'title': 'slow-start',
        'y1': 'Bytes',
        'y2': 'Time (ms)'
    },
    {
        'data': ['max_window', 'cur_window', 'our_delay', 'target_delay', 'ssthres'],
        'title': 'cwnd',
        'y1': 'Bytes',
        'y2': 'Time (ms)'
    },
    {
        'data': ['max_window', 'cur_window', 'packet_loss'],
        'title': 'packet-loss',
        'y1': 'Bytes',
        'y2': 'count'
    },
    {
        'data': ['max_window', 'cur_window', 'packet_timeout'],
        'title': 'packet-timeout',
        'y1': 'Bytes',
        'y2': 'count'
    },
    {
        'data': ['max_window', 'cur_window', 'bytes_sent', 'bytes_resent'],
        'title': 'cumulative-bytes-sent',
        'y1': 'Bytes',
        'y2': 'Cumulative Bytes'
    },
    {
        'data': ['max_window', 'cur_window', 'rto', 'timeout'],
        'title': 'connection-timeout',
        'y1': 'Bytes',
        'y2': 'Time (ms)'
    },
    {
        'data': ['our_delay', 'max_window', 'target_delay', 'cur_window', 'wnduser', 'cur_window_packets'],
        'title': 'uploading',
        'y1': 'Bytes',
        'y2': 'Time (ms)'
    },
    {
        'data': ['our_delay', 'max_window', 'target_delay', 'cur_window', 'send_buffer'],
        'title': 'uploading-packets',
        'y1': 'Bytes',
        'y2': 'Time (ms)'
    },
    {
        'data': ['their_delay', 'target_delay', 'rtt'],
        'title': 'their-delay',
        'y1': '',
        'y2': 'Time (ms)'
    },
    {
        'data': ['their_actual_delay', 'their_delay_base'],
        'title': 'their-delay-base',
        'y1': 'Time (us)',
        'y2': ''
    },
    {
        'data': ['our_delay', 'target_delay', 'rtt'],
        'title': 'our-delay',
        'y1': '',
        'y2': 'Time (ms)'
    },
    {
        'data': ['actual_delay', 'delay_base'],
        'title': 'our-delay-base',
        'y1': 'Time (us)',
        'y2': ''
    },
    {
        'data': ['ack_nr', 'seq_nr'],
        'title': 'sequence-numbers',
        'y1': 'Bytes',
        'y2': 'seqnr'
    },
    {
        'data': ['max_window', 'cur_window', 'acked_bytes'],
        'title': 'ack-rate',
        'y1': 'Bytes',
        'y2': 'Bytes'
    }
]

out = open('utp.gnuplot', 'w+')

files = ''

# print('set xtics 0, 20', file=out)
print("set term png size 1280,800", file=out)
print('set output "%s.delays.png"' % out_file, file=out)
print('set xrange [0:200]', file=out)
print('set xlabel "delay (ms)"', file=out)
print('set boxwidth 1', file=out)
print('set ylabel "number of packets"', file=out)
print('plot "%s.histogram" using 1:2 with boxes fs solid 0.3' % out_file, file=out)
files += out_file + '.delays.png '

print('set output "%s.packet_sizes.png"' % out_file, file=out)
print('set xrange [0:*]', file=out)
print('set xlabel "packet size (B)"', file=out)
print('set boxwidth 1', file=out)
print('set ylabel "number of packets sent"', file=out)
print('set logscale y', file=out)
print('plot "%s_packet_size.histogram" using 1:2 with boxes fs solid 0.3' % out_file, file=out)
print('set nologscale y', file=out)
files += out_file + '.packet_sizes.png '

print("set style data steps", file=out)
# print("set yrange [0:*]", file=out)
print("set y2range [*:*]", file=out)
# set hidden3d
# set title "Peer bandwidth distribution"
# set xlabel "Ratio"

for p in plot:
    print('set title "%s %s"' % (p['title'], title), file=out)
    print('set xlabel "time (s)"', file=out)
    print('set ylabel "%s"' % p['y1'], file=out)
    print("set tics nomirror", file=out)
    print('set y2tics', file=out)
    print('set y2label "%s"' % p['y2'], file=out)
    print('set xrange [0:*]', file=out)
    print("set key box", file=out)
    print("set term png size 1280,800", file=out)
    print('set output "%s-%s.png"' % (out_file, p['title']), file=out)
    files += '%s-%s.png ' % (out_file, p['title'])

    comma = ''
    print("plot", end=' ', file=out)

    for c in p['data']:
        if c not in metrics:
            continue
        i = columns.index(c)
        print('%s"%s" using ($1/1000):%d title "%s-%s" axes %s with %s' %
              (comma, out_file, i + 2, metrics[c][0], metrics[c][1], metrics[c][1], metrics[c][2]), end=' ', file=out)
        comma = ', '
    print('', file=out)

out.close()

os.system("gnuplot utp.gnuplot")

os.system("open %s" % files)



#!/usr/bin/env python3
# Copyright Arvid Norberg 2008. Use, modification and distribution is
# subject to the Boost Software License, Version 1.0. (See accompanying
# file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
from __future__ import print_function

import libtorrent as lt
import time
import sys

ses = lt.session({'listen_interfaces': '0.0.0.0:6881'})

info = lt.torrent_info(sys.argv[1])
h = ses.add_torrent({'ti': info, 'save_path': '.'})
s = h.status()
print('starting', s.name)

while (not s.is_seeding):
    s = h.status()

    print('\r%.2f%% complete (down: %.1f kB/s up: %.1f kB/s peers: %d) %s' % (
        s.progress * 100, s.download_rate / 1000, s.upload_rate / 1000,
        s.num_peers, s.state), end=' ')

    alerts = ses.pop_alerts()
    for a in alerts:
        if a.category() & lt.alert.category_t.error_notification:
            print(a)

    sys.stdout.flush()

    time.sleep(1)

print(h.status().name, 'complete')



#!/usr/bin/env python3
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4

# Copyright (c) 2016, Arvid Norberg
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#     * Neither the name of the author nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# this script can parse and generate reports from the alert log from a
# libtorrent session

import os
import sys
import math
from multiprocessing.pool import ThreadPool
from pathlib import Path
from argparse import ArgumentParser
from typing import Optional


line_graph = 0
histogram = 1
stacked = 2
diff = 3


def graph_colors() -> list[str]:
    colors: list[str] = []

    pattern = [[0, 0, 1], [0, 1, 0], [1, 0, 0], [1, 0, 1], [0, 1, 1], [1, 1, 0]]
    brightness = [0xD8, 0xBB, 0x60]

    for op in range(3):
        for c in pattern:
            c = [v * brightness[op] for v in c]
            colors.append("#%02x%02x%02x" % (c[0], c[1], c[2]))
    return colors


def gradient_colors(num_colors: int) -> list[str]:
    colors = []
    for i in range(0, num_colors):
        f = i / float(num_colors)
        pi = 3.1415927
        r = max(int(255 * (math.sin(f * pi) + 0.2)), 0)
        g = max(int(255 * (math.sin((f - 0.5) * pi) + 0.2)), 0)
        b = max(int(255 * (math.sin((f + 0.5) * pi) + 0.2)), 0)
        c = "#%02x%02x%02x" % (min(r, 255), min(g, 255), min(b, 255))
        colors.append(c)
    return colors


def plot_fun(script: Path) -> None:
    try:
        ret = os.system('gnuplot "%s" 2>/dev/null' % script)
    except Exception as e:
        print("please install gnuplot: sudo apt install gnuplot")
        raise e
    if ret != 0 and ret != 256:
        print("gnuplot failed: %d\n" % ret)
        raise Exception("abort")

    sys.stdout.write(".")
    sys.stdout.flush()


def to_title(key: str) -> str:
    return key.replace("_", " ").replace(".", " - ")


def gen_report(
    output_dir: Path,
    keys: list[str],
    name: str,
    unit: str,
    lines: list[str],
    short_unit: str,
    generation: int,
    log_file: Path,
    options: dict[str, int],
) -> Optional[Path]:
    filename = output_dir / f"{name}_{generation:04d}.png"
    thumb = output_dir / f"{name}_{generation:04d}_thumb.png"

    # don't re-render a graph unless the logfile has changed
    try:
        dst1 = filename.stat()
        dst2 = thumb.stat()
        src = log_file.stat()

        if dst1.st_mtime > src.st_mtime and dst2.st_mtime > src.st_mtime:
            sys.stdout.write(".")
            return None

    except Exception:
        pass

    script = output_dir / f"{name}_{generation:04d}.gnuplot"
    with open(script, "w") as out:
        print("set term png size 1200,700", file=out)
        print('set output "%s"' % filename, file=out)
        if "allow-negative" not in options:
            print("set yrange [0:*]", file=out)
        print("set tics nomirror", file=out)
        print("set key box", file=out)
        print("set key left top", file=out)

        colors = graph_colors()

        try:
            if "gradient" in options:
                colors = gradient_colors(options["gradient"])
        except Exception:
            pass

        first = True
        color = 0
        if options["type"] == histogram:
            binwidth = options["binwidth"]
            numbins = int(options["numbins"])

            print("binwidth=%f" % binwidth, file=out)
            print("set boxwidth binwidth", file=out)
            print("bin(x,width)=width*floor(x/width) + binwidth/2", file=out)
            print("set xrange [0:%f]" % (binwidth * numbins), file=out)
            print('set xlabel "%s"' % unit, file=out)
            print('set ylabel "number"', file=out)

            k = lines[0]
            try:
                column = keys.index(k) + 2
            except Exception:
                print('"%s" not found' % k)
                return None
            print(
                'plot "%s" using (bin($%d,binwidth)):(1.0) smooth freq with boxes'
                % (log_file, column),
                file=out,
            )
            print("", file=out)
            print("", file=out)
            print("", file=out)

        elif options["type"] == stacked:
            print("set xrange [0:*]", file=out)
            print('set ylabel "%s"' % unit, file=out)
            print('set xlabel "time (s)"', file=out)
            print('set format y "%%.1s%%c%s";' % short_unit, file=out)
            print("set style fill solid 1.0 noborder", file=out)
            print("plot", end=" ", file=out)
            graph = ""
            plot_expression = ""
            for k in lines:
                try:
                    column = keys.index(k) + 2
                except Exception:
                    print('"%s" not found' % k)
                    continue
                if not first:
                    plot_expression = ", " + plot_expression
                    graph += "+"
                axis = "x1y1"
                graph += "$%d" % column
                plot_expression = (
                    ' "%s" using 1:(%s) title "%s" axes %s with filledcurves x1 lc rgb "%s"'
                    % (log_file, graph, to_title(k), axis, colors[color % len(colors)])
                    + plot_expression
                )
                first = False
                color += 1
            print(plot_expression, file=out)
        elif options["type"] == diff:
            print("set xrange [0:*]", file=out)
            print('set ylabel "%s"' % unit, file=out)
            print('set xlabel "time (s)"', file=out)
            print('set format y "%%.1s%%c%s";' % short_unit, file=out)
            graph = ""
            title = ""
            for k in lines:
                try:
                    column = keys.index(k) + 2
                except Exception:
                    print('"%s" not found' % k)
                    continue
                if not first:
                    graph += "-"
                    title += " - "
                graph += "$%d" % column
                title += to_title(k)
                first = False
            print(
                'plot "%s" using 1:(%s) title "%s" with step'
                % (log_file, graph, title),
                file=out,
            )
        else:
            print("set xrange [0:*]", file=out)
            print('set ylabel "%s"' % unit, file=out)
            print('set xlabel "time (s)"', file=out)
            print('set format y "%%.1s%%c%s";' % short_unit, file=out)
            print("plot", end=" ", file=out)
            for k in lines:
                try:
                    column = keys.index(k) + 2
                except Exception:
                    print('"%s" not found' % k)
                    continue
                if not first:
                    print(", ", end=" ", file=out)
                axis = "x1y1"
                print(
                    ' "%s" using 1:%d title "%s" axes %s with steps lc rgb "%s"'
                    % (
                        log_file,
                        column,
                        to_title(k),
                        axis,
                        colors[color % len(colors)],
                    ),
                    end=" ",
                    file=out,
                )
                first = False
                color += 1
            print("", file=out)

        print("set term png size 150,100", file=out)
        print('set output "%s"' % thumb, file=out)
        print("set key off", file=out)
        print("unset tics", file=out)
        print('set format x ""', file=out)
        print('set format y ""', file=out)
        print('set xlabel ""', file=out)
        print('set ylabel ""', file=out)
        print('set y2label ""', file=out)
        print("set rmargin 0", file=out)
        print("set lmargin 0", file=out)
        print("set tmargin 0", file=out)
        print("set bmargin 0", file=out)
        print("replot", file=out)
    return script


def gen_html(
    reports: list[tuple[str, str, str, str, list[str], dict[str, int]]],
    num_generations: int,
    output_dir: Path,
) -> None:
    with open(output_dir / "index.html", "w+") as file:

        css = """img { margin: 0}
#head { display: block }
#graphs { white-space:nowrap; }
h1 { line-height: 1; display: inline }
h2 { line-height: 1; display: inline; font-size: 1em; font-weight: normal};"""

        print(
            '<html><head><style type="text/css">%s</style></head><body>' % css,
            file=file,
        )

        for i in reports:
            print(
                '<div id="head"><h1>%s </h1><h2>%s</h2><div><div id="graphs">'
                % (i[0], i[3]),
                file=file,
            )
            for g in range(num_generations):
                print(
                    '<a href="%s_%04d.png"><img src="%s_%04d_thumb.png"></a>'
                    % (i[0], g, i[0], g),
                    file=file,
                )
            print("</div>", file=file)

        print("</body></html>", file=file)


def main(input_file: Path, num_threads: int, output_dir: Path) -> None:
    thread_pool = ThreadPool(num_threads)
    output_dir.mkdir(parents=True, exist_ok=True)
    g = 0
    with open(input_file) as stat, open(output_dir / f"counters.{g:04}.dat", "w+") as data_out:
        line = stat.readline()
        print("looking for stats header")
        while "session stats header:" not in line:
            line = stat.readline()

        print("found")
        keys = line.split("session stats header:")[1].strip().split(", ")

        idx = 0
        for line in stat:
            if "session stats (" not in line:
                continue
            data_out.write(
                f"{idx}\t" + line.split(" values): ")[1].strip().replace(", ", "\t")
                + "\n"
            )
            idx += 1
            if idx >= 3600:
                g += 1
                data_out.close()
                data_out = open(output_dir / f"counters.{g:04}.dat", "w+")
                idx = 0

    num_generations = g + 1

    reports: list[tuple[str, str, str, str, list[str], dict[str, int]]] = [
        (
            "torrents",
            "num",
            "",
            "number of torrents in different torrent states",
            [
                "ses.num_downloading_torrents",
                "ses.num_seeding_torrents",
                "ses.num_checking_torrents",
                "ses.num_stopped_torrents",
                "ses.num_upload_only_torrents",
                "ses.num_error_torrents",
                "ses.num_queued_seeding_torrents",
                "ses.num_queued_download_torrents",
            ],
            {"type": stacked},
        ),
        (
            "peers",
            "num",
            "",
            "num connected peers",
            ["peer.num_peers_connected", "peer.num_peers_half_open"],
            {"type": stacked},
        ),
        (
            "peers_max",
            "num",
            "",
            "num connected peers",
            ["peer.num_peers_connected", "peer.num_peers_half_open"],
            {},
        ),
        (
            "peer_churn",
            "num",
            "",
            "connecting and disconnecting peers",
            [
                "peer.num_peers_half_open",
                "peer.connection_attempts",
                "peer.boost_connection_attempts",
                "peer.missed_connection_attempts",
                "peer.no_peer_connection_attempts",
            ],
            {},
        ),
        (
            "new_peers",
            "num",
            "",
            "",
            ["peer.incoming_connections", "peer.connection_attempts"],
            {},
        ),
        (
            "connection_attempts",
            "num",
            "",
            "",
            ["peer.connection_attempt_loops", "peer.connection_attempts"],
            {},
        ),
        (
            "pieces",
            "num",
            "",
            "number completed pieces",
            [
                "ses.num_total_pieces_added",
                "ses.num_piece_passed",
                "ses.num_piece_failed",
            ],
            {},
        ),
        (
            "disk_write_queue",
            "Bytes",
            "B",
            "bytes queued up by peers, to be written to disk",
            ["disk.queued_write_bytes"],
            {},
        ),
        (
            "peers_requests",
            "num",
            "",
            "incoming piece request rate",
            [
                "peer.piece_requests",
                "peer.max_piece_requests",
                "peer.invalid_piece_requests",
                "peer.choked_piece_requests",
                "peer.cancelled_piece_requests",
            ],
            {},
        ),
        (
            "peers_upload",
            "num",
            "",
            "number of peers by state wrt. uploading",
            [
                "peer.num_peers_up_disk",
                "peer.num_peers_up_interested",
                "peer.num_peers_up_unchoked_all",
                "peer.num_peers_up_unchoked_optimistic",
                "peer.num_peers_up_unchoked",
                "peer.num_peers_up_requests",
            ],
            {},
        ),
        (
            "peers_download",
            "num",
            "",
            "number of peers by state wrt. downloading",
            [
                "peer.num_peers_down_interested",
                "peer.num_peers_down_unchoked",
                "peer.num_peers_down_requests",
                "peer.num_peers_down_disk",
            ],
            {},
        ),
        (
            "peer_errors",
            "num",
            "",
            "number of peers by error that disconnected them",
            [
                "peer.disconnected_peers",
                "peer.eof_peers",
                "peer.connreset_peers",
                "peer.connrefused_peers",
                "peer.connaborted_peers",
                "peer.perm_peers",
                "peer.buffer_peers",
                "peer.unreachable_peers",
                "peer.broken_pipe_peers",
                "peer.addrinuse_peers",
                "peer.no_access_peers",
                "peer.invalid_arg_peers",
                "peer.aborted_peers",
            ],
            {"type": stacked},
        ),
        (
            "peer_errors_incoming",
            "num",
            "",
            "number of peers by incoming or outgoing connection",
            ["peer.error_incoming_peers", "peer.error_outgoing_peers"],
            {},
        ),
        (
            "peer_errors_transport",
            "num",
            "",
            "number of peers by transport protocol",
            ["peer.error_tcp_peers", "peer.error_utp_peers"],
            {},
        ),
        (
            "peer_errors_encryption",
            "num",
            "",
            "number of peers by encryption level",
            [
                "peer.error_encrypted_peers",
                "peer.error_rc4_peers",
            ],
            {},
        ),
        (
            "incoming requests",
            "num",
            "",
            "incoming 16kiB block requests",
            ["ses.num_incoming_request"],
            {},
        ),
        (
            "waste",
            "downloaded bytes",
            "B",
            "proportion of all downloaded bytes that were wasted",
            [
                "net.recv_failed_bytes",
                "net.recv_redundant_bytes",
                "net.recv_ip_overhead_bytes",
            ],
            {"type": stacked},
        ),
        (
            "waste by source",
            "num wasted bytes",
            "B",
            "what is causing the waste",
            [
                "ses.waste_piece_timed_out",
                "ses.waste_piece_cancelled",
                "ses.waste_piece_unknown",
                "ses.waste_piece_seed",
                "ses.waste_piece_end_game",
                "ses.waste_piece_closing",
            ],
            {"type": stacked},
        ),
        (
            "disk_time",
            "% of total disk job time",
            "%%",
            "proportion of time spent by the disk thread",
            ["disk.disk_read_time", "disk.disk_write_time", "disk.disk_hash_time"],
            {"type": stacked},
        ),
        (
            "disk_queue",
            "blocks (16kiB)",
            "",
            "disk store-buffer size",
            [
                "disk.num_write_jobs",
                "disk.num_read_jobs",
                "disk.num_jobs",
                "disk.queued_disk_jobs",
                "disk.blocked_disk_jobs",
            ],
            {},
        ),
        (
            "disk fences",
            "num",
            "",
            "number of jobs currently blocked by a fence job",
            ["disk.blocked_disk_jobs"],
            {},
        ),
        (
            "fence jobs",
            "num",
            "",
            "active fence jobs per type",
            [
                "disk.num_fenced_move_storage",
                "disk.num_fenced_release_files",
                "disk.num_fenced_delete_files",
                "disk.num_fenced_check_fastresume",
                "disk.num_fenced_save_resume_data",
                "disk.num_fenced_rename_file",
                "disk.num_fenced_stop_torrent",
                "disk.num_fenced_file_priority",
                "disk.num_fenced_clear_piece",
            ],
            {"type": stacked},
        ),
        (
            "disk threads",
            "num",
            "",
            "number of disk threads currently writing",
            ["disk.num_writing_threads", "disk.num_running_threads"],
            {},
        ),
        (
            "connection_type",
            "num",
            "",
            "peers by transport protocol",
            [
                "peer.num_tcp_peers",
                "peer.num_socks5_peers",
                "peer.num_http_proxy_peers",
                "peer.num_utp_peers",
                "peer.num_i2p_peers",
                "peer.num_ssl_peers",
                "peer.num_ssl_socks5_peers",
                "peer.num_ssl_http_proxy_peers",
                "peer.num_ssl_utp_peers",
            ],
            {},
        ),
        # (
        #     "uTP delay",
        #     "buffering delay",
        #     "s",
        #     "network delays measured by uTP",
        #     [
        #         "uTP peak send delay",
        #         "uTP peak recv delay",
        #         "uTP avg send delay",
        #         "uTP avg recv delay",
        #     ],
        # ),
        # (
        #     "uTP send delay histogram",
        #     "buffering delay",
        #     "s",
        #     "send delays measured by uTP",
        #     ["uTP avg send delay"],
        #     {"type": histogram, "binwidth": 0.05, "numbins": 100},
        # ),
        # (
        #     "uTP recv delay histogram",
        #     "buffering delay",
        #     "s",
        #     "receive delays measured by uTP",
        #     ["uTP avg recv delay"],
        #     {"type": histogram, "binwidth": 0.05, "numbins": 100},
        # ),
        (
            "uTP stats",
            "num",
            "",
            "number of uTP events",
            [
                "utp.utp_packet_loss",
                "utp.utp_timeout",
                "utp.utp_packets_in",
                "utp.utp_packets_out",
                "utp.utp_fast_retransmit",
                "utp.utp_packet_resend",
                "utp.utp_samples_above_target",
                "utp.utp_samples_below_target",
                "utp.utp_payload_pkts_in",
                "utp.utp_payload_pkts_out",
                "utp.utp_invalid_pkts_in",
                "utp.utp_redundant_pkts_in",
            ],
            {"type": stacked},
        ),
        (
            "boost.asio messages",
            "num events",
            "",
            "number of messages posted",
            [
                "net.on_read_counter",
                "net.on_write_counter",
                "net.on_tick_counter",
                "net.on_lsd_counter",
                "net.on_lsd_peer_counter",
                "net.on_udp_counter",
                "net.on_accept_counter",
                "net.on_disk_counter",
            ],
            {"type": stacked},
        ),
        (
            "send_buffer_sizes",
            "num",
            "",
            "",
            [
                "sock_bufs.socket_send_size3",
                "sock_bufs.socket_send_size4",
                "sock_bufs.socket_send_size5",
                "sock_bufs.socket_send_size6",
                "sock_bufs.socket_send_size7",
                "sock_bufs.socket_send_size8",
                "sock_bufs.socket_send_size9",
                "sock_bufs.socket_send_size10",
                "sock_bufs.socket_send_size11",
                "sock_bufs.socket_send_size12",
                "sock_bufs.socket_send_size13",
                "sock_bufs.socket_send_size14",
                "sock_bufs.socket_send_size15",
                "sock_bufs.socket_send_size16",
                "sock_bufs.socket_send_size17",
                "sock_bufs.socket_send_size18",
                "sock_bufs.socket_send_size19",
                "sock_bufs.socket_send_size20",
            ],
            {"type": stacked, "gradient": 18},
        ),
        (
            "recv_buffer_sizes",
            "num",
            "",
            "",
            [
                "sock_bufs.socket_recv_size3",
                "sock_bufs.socket_recv_size4",
                "sock_bufs.socket_recv_size5",
                "sock_bufs.socket_recv_size6",
                "sock_bufs.socket_recv_size7",
                "sock_bufs.socket_recv_size8",
                "sock_bufs.socket_recv_size9",
                "sock_bufs.socket_recv_size10",
                "sock_bufs.socket_recv_size11",
                "sock_bufs.socket_recv_size12",
                "sock_bufs.socket_recv_size13",
                "sock_bufs.socket_recv_size14",
                "sock_bufs.socket_recv_size15",
                "sock_bufs.socket_recv_size16",
                "sock_bufs.socket_recv_size17",
                "sock_bufs.socket_recv_size18",
                "sock_bufs.socket_recv_size19",
                "sock_bufs.socket_recv_size20",
            ],
            {"type": stacked, "gradient": 18},
        ),
        (
            "request latency",
            "us",
            "",
            "latency from receiving requests to sending response",
            ["disk.request_latency"],
            {},
        ),
        (
            "incoming messages",
            "num",
            "",
            "number of received bittorrent messages, by type",
            [
                "ses.num_incoming_choke",
                "ses.num_incoming_unchoke",
                "ses.num_incoming_interested",
                "ses.num_incoming_not_interested",
                "ses.num_incoming_have",
                "ses.num_incoming_bitfield",
                "ses.num_incoming_request",
                "ses.num_incoming_piece",
                "ses.num_incoming_cancel",
                "ses.num_incoming_dht_port",
                "ses.num_incoming_suggest",
                "ses.num_incoming_have_all",
                "ses.num_incoming_have_none",
                "ses.num_incoming_reject",
                "ses.num_incoming_allowed_fast",
                "ses.num_incoming_ext_handshake",
                "ses.num_incoming_pex",
                "ses.num_incoming_metadata",
                "ses.num_incoming_extended",
            ],
            {"type": stacked},
        ),
        (
            "outgoing messages",
            "num",
            "",
            "number of sent bittorrent messages, by type",
            [
                "ses.num_outgoing_choke",
                "ses.num_outgoing_unchoke",
                "ses.num_outgoing_interested",
                "ses.num_outgoing_not_interested",
                "ses.num_outgoing_have",
                "ses.num_outgoing_bitfield",
                "ses.num_outgoing_request",
                "ses.num_outgoing_piece",
                "ses.num_outgoing_cancel",
                "ses.num_outgoing_dht_port",
                "ses.num_outgoing_suggest",
                "ses.num_outgoing_have_all",
                "ses.num_outgoing_have_none",
                "ses.num_outgoing_reject",
                "ses.num_outgoing_allowed_fast",
                "ses.num_outgoing_ext_handshake",
                "ses.num_outgoing_pex",
                "ses.num_outgoing_metadata",
                "ses.num_outgoing_extended",
            ],
            {"type": stacked},
        ),
        (
            "request in balance",
            "num",
            "",
            "request and piece message balance",
            [
                "ses.num_incoming_request",
                "ses.num_outgoing_piece",
                "ses.num_outgoing_reject",
            ],
            {"type": diff},
        ),
        (
            "request out balance",
            "num",
            "",
            "request and piece message balance",
            [
                "ses.num_outgoing_request",
                "ses.num_incoming_piece",
                "ses.num_incoming_reject",
            ],
            {"type": diff},
        ),
        (
            "piece_picker_invocations",
            "invocations of piece picker",
            "",
            "",
            [
                "picker.reject_piece_picks",
                "picker.unchoke_piece_picks",
                "picker.incoming_redundant_piece_picks",
                "picker.incoming_piece_picks",
                "picker.end_game_piece_picks",
                "picker.snubbed_piece_picks",
                "picker.interesting_piece_picks",
                "picker.hash_fail_piece_picks",
            ],
            {"type": stacked},
        ),
        (
            "piece_picker_loops",
            "loops through piece picker",
            "",
            "",
            [
                "picker.piece_picker_partial_loops",
                "picker.piece_picker_suggest_loops",
                "picker.piece_picker_sequential_loops",
                "picker.piece_picker_reverse_rare_loops",
                "picker.piece_picker_rare_loops",
                "picker.piece_picker_rand_start_loops",
                "picker.piece_picker_rand_loops",
                "picker.piece_picker_busy_loops",
            ],
            {"type": stacked},
        ),
        (
            "async_accept",
            "number of outstanding accept calls",
            "",
            "",
            ["ses.num_outstanding_accept"],
            {},
        ),
        (
            "queued_trackers",
            "number of queued tracker announces",
            "",
            "",
            ["tracker.num_queued_tracker_announces"],
            {},
        ),
        (
            "file_pool_size",
            "file pool sze",
            "",
            "",
            ["disk.file_pool_size"],
            {},
        ),
        (
            "file_pool_misses",
            "file pool cache misses",
            "",
            "",
            ["disk.file_pool_misses", "disk.file_pool_thread_stall", "disk.file_pool_race"],
            {},
        ),
        # (
        #     "picker_full_partials_distribution",
        #     "full pieces",
        #     "",
        #     "",
        #     ["num full partial pieces"],
        #     {"type": histogram, "binwidth": 5, "numbins": 120},
        # ),
        # (
        #     "picker_partials_distribution",
        #     "partial pieces",
        #     "",
        #     "",
        #     ["num downloading partial pieces"],
        #     {"type": histogram, "binwidth": 5, "numbins": 120},
        # ),
    ]

    print("generating graphs")

    scripts = []
    g = 0

    for g in range(num_generations):
        scripts: list[Path] = []
        print(f"[{' ' * len(reports)}] {g:04}\r[", end="", flush=True)
        for rep in reports:
            try:
                options = rep[5]
            except Exception:
                options = {}
            if "type" not in options:
                options["type"] = line_graph

            script = gen_report(
                output_dir,
                keys,
                rep[0],
                rep[1],
                rep[4],
                rep[2],
                g,
                output_dir / f"counters.{g:04}.dat",
                options,
            )
            if script is not None:
                scripts.append(script)

        # run gnuplot on all scripts, in parallel
        thread_pool.map(plot_fun, scripts)
        scripts = []
        print()

    print("\ngenerating html")
    gen_html(reports, num_generations, output_dir)


if __name__ == "__main__":
    p = ArgumentParser()
    p.add_argument(
        "input",
        type=Path,
        help="libtorrent log file to parse. It must include the session_stats log entries",
    )
    p.add_argument(
        "--threads", default=8, type=int, help="The number of threads to run gnuplot in"
    )
    p.add_argument(
        "--output-dir",
        default="./session_stats_report",
        type=Path,
        help="The directory to save the output files in",
    )

    args = p.parse_args()
    main(args.input, args.threads, args.output_dir)



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
from __future__ import print_function

import urllib.parse
import glob
import os
import sys

verbose = '--verbose' in sys.argv
dump = '--dump' in sys.argv
internal = '--internal' in sys.argv
plain_output = '--plain-output' in sys.argv
single_page_output = '--single-page' in sys.argv
if plain_output:
    plain_file = open('plain_text_out.txt', 'w+')
in_code = None

paths = ['include/libtorrent/*.hpp', 'include/libtorrent/kademlia/*.hpp', 'include/libtorrent/extensions/*.hpp']

if internal:
    paths.append('include/libtorrent/aux_/*.hpp')

files = []

for p in paths:
    files.extend(glob.glob(os.path.join('..', p)))

functions = []
classes = []
enums = []
constants = {}

# maps filename to overview description
overviews = {}

# maps names -> URL
symbols = {}

# some files that need pre-processing to turn symbols into
# links into the reference documentation
preprocess_rst = \
    {
        'manual.rst': 'manual-ref.rst',
        'tuning.rst': 'tuning-ref.rst',
        'tutorial.rst': 'tutorial-ref.rst',
        'features.rst': 'features-ref.rst',
        'upgrade_to_1.2.rst': 'upgrade_to_1.2-ref.rst',
        'upgrade_to_2.0.rst': 'upgrade_to_2.0-ref.rst',
        'settings.rst': 'settings-ref.rst'
    }

# some pre-defined sections from the main manual
symbols = \
    {
        "queuing_": "manual-ref.html#queuing",
        "fast-resume_": "manual-ref.html#fast-resume",
        "storage-allocation_": "manual-ref.html#storage-allocation",
        "alerts_": "manual-ref.html#alerts",
        "upnp-and-nat-pmp_": "manual-ref.html#upnp-and-nat-pmp",
        "http-seeding_": "manual-ref.html#http-seeding",
        "metadata-from-peers_": "manual-ref.html#metadata-from-peers",
        "magnet-links_": "manual-ref.html#magnet-links",
        "ssl-torrents_": "manual-ref.html#ssl-torrents",
        "dynamic-loading-of-torrent-files_": "manual-ref.html#dynamic-loading-of-torrent-files",
        "session-statistics_": "manual-ref.html#session-statistics",
        "peer-classes_": "manual-ref.html#peer-classes",
        "BitTorrent-v2-torrents_": "manual-ref.html#bittorrent-v2-torrents",
    }

# parse out names of settings, and add them to the symbols list, to get cross
# references working
with open('../src/settings_pack.cpp') as f:
    for line in f:
        line = line.strip()
        if not line.startswith('SET('):
            continue

        name = line.split('(')[1].split(',')[0]
        symbols['settings_pack::' + name] = 'reference-Settings.html#' + name

static_links = \
    {
        ".. _`BEP 3`: https://www.bittorrent.org/beps/bep_0003.html",
        ".. _`BEP 17`: https://www.bittorrent.org/beps/bep_0017.html",
        ".. _`BEP 19`: https://www.bittorrent.org/beps/bep_0019.html",
        ".. _`BEP 38`: https://www.bittorrent.org/beps/bep_0038.html",
        ".. _`BEP 42`: https://www.bittorrent.org/beps/bep_0042.html",
        ".. _`BEP 52`: https://www.bittorrent.org/beps/bep_0052.html",
        ".. _`rate based choking`: manual-ref.html#rate-based-choking",
        ".. _extensions: manual-ref.html#extensions",
    }

anon_index = 0

category_mapping = {
    'ed25519.hpp': 'ed25519',
    'session.hpp': 'Session',
    'session_handle.hpp': 'Session',
    'torrent_handle.hpp': 'Torrent Handle',
    'torrent_info.hpp': 'Torrent Info',
    'announce_entry.hpp': 'Trackers',
    'peer_class_type_filter.hpp': 'PeerClass',
    'peer_class.hpp': 'PeerClass',
    'torrent_status.hpp': 'Torrent Status',
    'session_stats.hpp': 'Stats',
    'performance_counters.hpp': 'Stats',
    'read_resume_data.hpp': 'Resume Data',
    'write_resume_data.hpp': 'Resume Data',
    'add_torrent_params.hpp': 'Add Torrent',
    'client_data.hpp': 'Add Torrent',
    'session_status.hpp': 'Session',
    'session_params.hpp': 'Session',
    'error_code.hpp': 'Error Codes',
    'storage_defs.hpp': 'Storage',
    'file_storage.hpp': 'Storage',
    'disk_interface.hpp': 'Custom Storage',
    'disk_observer.hpp': 'Custom Storage',
    'mmap_disk_io.hpp': 'Storage',
    'disabled_disk_io.hpp': 'Storage',
    'posix_disk_io.hpp': 'Storage',
    'extensions.hpp': 'Plugins',
    'ut_metadata.hpp': 'Plugins',
    'ut_pex.hpp': 'Plugins',
    'ut_trackers.hpp': 'Plugins',
    'smart_ban.hpp': 'Plugins',
    'peer_connection_handle.hpp': 'Plugins',
    'create_torrent.hpp': 'Create Torrents',
    'alert.hpp': 'Alerts',
    'alert_types.hpp': 'Alerts',
    'bencode.hpp': 'Bencoding',
    'bdecode.hpp': 'Bdecoding',
    'entry.hpp': 'Bencoding',
    'time.hpp': 'Time',
    'escape_string.hpp': 'Utility',
    'enum_net.hpp': 'Network',
    'socket.hpp': 'Network',
    'address.hpp': 'Network',
    'socket_io.hpp': 'Network',
    'bitfield.hpp': 'Utility',
    'sha1_hash.hpp': 'Utility',
    'hasher.hpp': 'Utility',
    'identify_client.hpp': 'Utility',
    'ip_filter.hpp': 'Filter',
    'session_settings.hpp': 'Settings',
    'settings_pack.hpp': 'Settings',
    'fingerprint.hpp': 'Settings',
    'operations.hpp': 'Alerts',
    'disk_buffer_holder.hpp': 'Custom Storage',
    'alert_dispatcher.hpp': 'Alerts',
}

category_fun_mapping = {
    'min_memory_usage()': 'Settings',
    'high_performance_seed()': 'Settings',
    'default_disk_io_constructor()': 'Storage',
    'settings_interface': 'Custom Storage',
}


def categorize_symbol(name, filename):
    f = os.path.split(filename)[1]

    if name.endswith('_category()') \
            or name.endswith('_error_code') \
            or name.endswith('error_code_enum') \
            or name.endswith('errors'):
        return 'Error Codes'

    if name in category_fun_mapping:
        return category_fun_mapping[name]

    if f in category_mapping:
        return category_mapping[f]

    if filename.startswith('libtorrent/kademlia/'):
        return 'DHT'

    return 'Core'


def suppress_warning(filename, name):
    f = os.path.split(filename)[1]
    if f != 'alert_types.hpp':
        return False

    # if name.endswith('_alert') or name == 'message()':
    return True

    # return False


def first_item(itr):
    for i in itr:
        return i
    return None


def is_visible(desc):
    if desc.strip().startswith('hidden'):
        return False
    if internal:
        return True
    if desc.strip().startswith('internal'):
        return False
    return True


def highlight_signature(s):
    name = s.split('(', 1)
    name2 = name[0].split(' ')
    if len(name2[-1]) == 0:
        return s

    # make the name of the function bold
    name2[-1] = '**' + name2[-1] + '** '

    # if there is a return value, make sure we preserve pointer types
    if len(name2) > 1:
        name2[0] = name2[0].replace('*', '\\*')
    name[0] = ' '.join(name2)

    # we have to escape asterisks, since this is rendered into
    # a parsed literal in rst
    name[1] = name[1].replace('*', '\\*')

    # we also have to escape colons
    name[1] = name[1].replace(':', '\\:')

    # escape trailing underscores
    name[1] = name[1].replace('_', '\\_')

    # comments in signatures are italic
    name[1] = name[1].replace('/\\*', '*/\\*')
    name[1] = name[1].replace('\\*/', '\\*/*')
    return '('.join(name)


def highlight_name(s):
    if '=' in s:
        splitter = ' = '
    elif '{' in s:
        splitter = '{'
    else:
        return s

    name = s.split(splitter, 1)
    name2 = name[0].split(' ')
    if len(name2[-1]) == 0:
        return s

    name2[-1] = '**' + name2[-1] + '** '
    name[0] = ' '.join(name2)
    return splitter.join(name)


def html_sanitize(s):
    ret = ''
    for i in s:
        if i == '<':
            ret += '&lt;'
        elif i == '>':
            ret += '&gt;'
        elif i == '&':
            ret += '&amp;'
        else:
            ret += i
    return ret


def looks_like_namespace(line):
    line = line.strip()
    if line.startswith('namespace'):
        return True
    return False


def looks_like_blank(line):
    line = line.split('//')[0]
    line = line.replace('{', '')
    line = line.replace('}', '')
    line = line.replace('[', '')
    line = line.replace(']', '')
    line = line.replace(';', '')
    line = line.strip()
    return len(line) == 0


def looks_like_variable(line):
    line = line.split('//')[0]
    line = line.strip()
    if ' ' not in line and '\t' not in line:
        return False
    if line.startswith('friend '):
        return False
    if line.startswith('enum '):
        return False
    if line.startswith(','):
        return False
    if line.startswith(':'):
        return False
    if line.startswith('typedef'):
        return False
    if line.startswith('using'):
        return False
    if ' = ' in line:
        return True
    if line.endswith(';'):
        return True
    return False


def looks_like_constant(line):
    line = line.strip()
    if line.startswith('inline'):
        line = line.split('inline')[1]
    line = line.strip()
    if not line.startswith('constexpr'):
        return False
    line = line.split('constexpr')[1]
    return looks_like_variable(line)


def looks_like_forward_decl(line):
    line = line.split('//')[0]
    line = line.strip()
    if not line.endswith(';'):
        return False
    if '{' in line:
        return False
    if '}' in line:
        return False
    if line.startswith('friend '):
        return True
    if line.startswith('struct '):
        return True
    if line.startswith('class '):
        return True
    return False


def looks_like_function(line):
    line = line.split('//')[0]
    if line.startswith('friend class '):
        return False
    if line.startswith('friend struct '):
        return False
    if '::' in line.split('(')[0].split(' ')[-1]:
        return False
    if line.startswith(','):
        return False
    if line.startswith(':'):
        return False
    return '(' in line


def parse_function(lno, lines, filename):

    start_paren = 0
    end_paren = 0
    signature = ''

    global orphaned_export
    orphaned_export = False

    while lno < len(lines):
        line = lines[lno].strip()
        lno += 1
        if line.startswith('//'):
            continue

        start_paren += line.count('(')
        end_paren += line.count(')')

        sig_line = line.replace('TORRENT_EXPORT ', '') \
            .replace('TORRENT_EXTRA_EXPORT', '') \
            .replace('TORRENT_V3_EXPLICIT', '') \
            .replace('TORRENT_COUNTER_NOEXCEPT', '') \
            .split('//')[0].strip()
        if signature != '':
            sig_line = '\n   ' + sig_line
        signature += sig_line
        if verbose:
            print('fun     %s' % line)

        if start_paren > 0 and start_paren == end_paren:
            if signature[-1] != ';':
                # we also need to consume the function body
                start_paren = 0
                end_paren = 0
                for i in range(len(signature)):
                    if signature[i] == '(':
                        start_paren += 1
                    elif signature[i] == ')':
                        end_paren += 1

                    if start_paren > 0 and start_paren == end_paren:
                        for k in range(i, len(signature)):
                            if signature[k] == ':' or signature[k] == '{':
                                signature = signature[0:k].strip()
                                break
                        break

                lno = consume_block(lno - 1, lines)
                signature += ';'
            ret = [{'file': filename[11:], 'signatures': set([signature]), 'names': set(
                [signature.split('(')[0].split(' ')[-1].strip() + '()'])}, lno]
            if first_item(ret[0]['names']) == '()':
                return [None, lno]
            return ret
    if len(signature) > 0:
        print('\x1b[31mFAILED TO PARSE FUNCTION\x1b[0m %s\nline: %d\nfile: %s' % (signature, lno, filename))
    return [None, lno]


def add_desc(line):
    # plain output prints just descriptions and filters out c++ code.
    # it's used to run spell checker over
    if plain_output:
        for s in line.split('\n'):
            # if the first character is a space, strip it
            if len(s) > 0 and s[0] == ' ':
                s = s[1:]
            global in_code
            if in_code is not None and not s.startswith(in_code) and len(s) > 1:
                in_code = None

            if s.strip().startswith('.. code::'):
                in_code = s.split('.. code::')[0] + '\t'

            # strip out C++ code from the plain text output since it's meant for
            # running spell checking over
            if not s.strip().startswith('.. ') and in_code is None:
                plain_file.write(s + '\n')


def parse_class(lno, lines, filename):
    start_brace = 0
    end_brace = 0

    name = ''
    funs = []
    fields = []
    enums = []
    state = 'public'
    context = ''
    class_type = 'struct'
    blanks = 0
    decl = ''

    while lno < len(lines):
        line = lines[lno].strip()
        decl += lines[lno].replace('TORRENT_EXPORT ', '') \
            .replace('TORRENT_EXTRA_EXPORT', '') \
            .replace('TORRENT_V3_EXPLICIT', '') \
            .replace('TORRENT_COUNTER_NOEXCEPT', '').split('{')[0].strip()
        if '{' in line:
            break
        if verbose:
            print('class  %s' % line)
        lno += 1

    if decl.startswith('class'):
        state = 'private'
        class_type = 'class'

    name = decl.split(':')[0].replace('class ', '').replace('struct ', '').replace('final', '').strip()

    global orphaned_export
    orphaned_export = False

    while lno < len(lines):
        line = lines[lno].strip()
        lno += 1

        if line == '':
            blanks += 1
            context = ''
            continue

        if line.startswith('/*'):
            lno = consume_comment(lno - 1, lines)
            continue

        if line.startswith('#'):
            lno = consume_ifdef(lno - 1, lines, True)
            continue

        if 'TORRENT_DEFINE_ALERT' in line:
            if verbose:
                print('xx    %s' % line)
            blanks += 1
            continue
        if 'TORRENT_DEPRECATED' in line:
            if verbose:
                print('xx    %s' % line)
            blanks += 1
            continue

        if line.startswith('//'):
            if verbose:
                print('desc  %s' % line)

            line = line[2:]
            if len(line) and line[0] == ' ':
                line = line[1:]
            context += line + '\n'
            continue

        start_brace += line.count('{')
        end_brace += line.count('}')

        if line == 'private:':
            state = 'private'
        elif line == 'protected:':
            state = 'protected'
        elif line == 'public:':
            state = 'public'

        if start_brace > 0 and start_brace == end_brace:
            return [{'file': filename[11:], 'enums': enums, 'fields':fields,
                     'type': class_type, 'name': name, 'decl': decl, 'fun': funs}, lno]

        if state != 'public' and not internal:
            if verbose:
                print('private %s' % line)
            blanks += 1
            continue

        if start_brace - end_brace > 1:
            if verbose:
                print('scope   %s' % line)
            blanks += 1
            continue

        if looks_like_function(line):
            current_fun, lno = parse_function(lno - 1, lines, filename)
            if current_fun is not None and is_visible(context):
                if context == '' and blanks == 0 and len(funs):
                    funs[-1]['signatures'].update(current_fun['signatures'])
                    funs[-1]['names'].update(current_fun['names'])
                else:
                    if 'TODO: ' in context:
                        print('TODO comment in public documentation: %s:%d' % (filename, lno))
                        sys.exit(1)
                    current_fun['desc'] = context
                    add_desc(context)
                    if context == '' and not suppress_warning(filename, first_item(current_fun['names'])):
                        print('WARNING: member function "%s" is not documented: \x1b[34m%s:%d\x1b[0m'
                              % (name + '::' + first_item(current_fun['names']), filename, lno))
                    funs.append(current_fun)
                context = ''
                blanks = 0
            continue

        if looks_like_variable(line):
            if 'constexpr static' in line:
                print('ERROR: found "constexpr static", use "static constexpr" instead for consistency!\n%s:%d\n%s'
                      % (filename, lno, line))
                sys.exit(1)
            if verbose:
                print('var     %s' % line)
            if not is_visible(context):
                continue
            line = line.split('//')[0].strip()
            # the name may look like this:
            # std::uint8_t fails : 7;
            # int scrape_downloaded = -1;
            # static constexpr peer_flags_t interesting{0x1};
            n = line.split('=')[0].split('{')[0].strip().split(' : ')[0].split(' ')[-1].split(':')[0].split(';')[0]
            if context == '' and blanks == 0 and len(fields):
                fields[-1]['names'].append(n)
                fields[-1]['signatures'].append(line)
            else:
                if context == '' and not suppress_warning(filename, n):
                    print('WARNING: field "%s" is not documented: \x1b[34m%s:%d\x1b[0m'
                          % (name + '::' + n, filename, lno))
                add_desc(context)
                fields.append({'signatures': [line], 'names': [n], 'desc': context})
            context = ''
            blanks = 0
            continue

        if line.startswith('enum '):
            if verbose:
                print('enum    %s' % line)
            if not is_visible(context):
                consume_block(lno - 1, lines)
            else:
                enum, lno = parse_enum(lno - 1, lines, filename)
                if enum is not None:
                    if 'TODO: ' in context:
                        print('TODO comment in public documentation: %s:%d' % (filename, lno))
                        sys.exit(1)
                    enum['desc'] = context
                    add_desc(context)
                    if context == '' and not suppress_warning(filename, enum['name']):
                        print('WARNING: enum "%s" is not documented: \x1b[34m%s:%d\x1b[0m'
                              % (name + '::' + enum['name'], filename, lno))
                    enums.append(enum)
                context = ''
            continue

        context = ''

        if verbose:
            if looks_like_forward_decl(line) \
                    or looks_like_blank(line) \
                    or looks_like_namespace(line):
                print('--      %s' % line)
            else:
                print('??      %s' % line)

    if len(name) > 0:
        print('\x1b[31mFAILED TO PARSE CLASS\x1b[0m %s\nfile: %s:%d' % (name, filename, lno))
    return [None, lno]


def parse_constant(lno, lines, filename):
    line = lines[lno].strip()
    if verbose:
        print('const   %s' % line)
    line = line.split('=')[0]
    if 'constexpr' in line:
        line = line.split('constexpr')[1]
    if '{' in line and '}' in line:
        line = line.split('{')[0]
    t, name = line.strip().rsplit(' ', 1)
    return [{'file': filename[11:], 'type': t, 'name': name}, lno + 1]


def parse_enum(lno, lines, filename):
    start_brace = 0
    end_brace = 0
    global anon_index

    line = lines[lno].strip()
    name = line.replace('enum ', '').replace('class ', '').split(':')[0].split('{')[0].strip()
    if len(name) == 0:
        if not internal:
            print('WARNING: anonymous enum at: \x1b[34m%s:%d\x1b[0m' % (filename, lno))
            lno = consume_block(lno - 1, lines)
            return [None, lno]
        name = 'anonymous_enum_%d' % anon_index
        anon_index += 1

    values = []
    context = ''
    if '{' not in line:
        if verbose:
            print('enum  %s' % lines[lno])
        lno += 1

    val = 0
    while lno < len(lines):
        line = lines[lno].strip()
        lno += 1

        if line.startswith('//'):
            if verbose:
                print('desc  %s' % line)
            line = line[2:]
            if len(line) and line[0] == ' ':
                line = line[1:]
            context += line + '\n'
            continue

        if line.startswith('#'):
            lno = consume_ifdef(lno - 1, lines)
            continue

        start_brace += line.count('{')
        end_brace += line.count('}')

        if '{' in line:
            line = line.split('{')[1]
        line = line.split('}')[0]

        if len(line):
            if verbose:
                print('enumv %s' % lines[lno - 1])
            for v in line.split(','):
                v = v.strip()
                if v.startswith('//'):
                    break
                if v == '':
                    continue
                valstr = ''
                try:
                    if '=' in v:
                        val = int(v.split('=')[1].strip(), 0)
                    valstr = str(val)
                except Exception:
                    pass

                if '=' in v:
                    v = v.split('=')[0].strip()
                if is_visible(context):
                    add_desc(context)
                    values.append({'name': v.strip(), 'desc': context, 'val': valstr})
                    if verbose:
                        print('enumv %s' % valstr)
                context = ''
                val += 1
        else:
            if verbose:
                print('??    %s' % lines[lno - 1])

        if start_brace > 0 and start_brace == end_brace:
            return [{'file': filename[11:], 'name': name, 'values': values}, lno]

    if len(name) > 0:
        print('\x1b[31mFAILED TO PARSE ENUM\x1b[0m %s\nline: %d\nfile: %s' % (name, lno, filename))
    return [None, lno]


def consume_block(lno, lines):
    start_brace = 0
    end_brace = 0

    while lno < len(lines):
        line = lines[lno].strip()
        if verbose:
            print('xx    %s' % line)
        lno += 1

        start_brace += line.count('{')
        end_brace += line.count('}')

        if start_brace > 0 and start_brace == end_brace:
            break
    return lno


def consume_comment(lno, lines):
    while lno < len(lines):
        line = lines[lno].strip()
        if verbose:
            print('xx    %s' % line)
        lno += 1
        if '*/' in line:
            break

    return lno


def trim_define(line):
    return line.replace('#ifndef', '').replace('#ifdef', '') \
            .replace('#if', '').replace('defined', '') \
            .replace('TORRENT_ABI_VERSION == 1', '') \
            .replace('TORRENT_ABI_VERSION <= 2', '') \
            .replace('TORRENT_ABI_VERSION < 3', '') \
            .replace('||', '').replace('&&', '').replace('(', '').replace(')', '') \
            .replace('!', '').replace('\\', '').strip()


def consume_ifdef(lno, lines, warn_on_ifdefs=False):
    line = lines[lno].strip()
    lno += 1

    start_if = 1
    end_if = 0

    if verbose:
        print('prep  %s' % line)

    if warn_on_ifdefs and line.strip().startswith('#if'):
        while line.endswith('\\'):
            lno += 1
            line += lines[lno].strip()
            if verbose:
                print('prep  %s' % lines[lno].trim())
        define = trim_define(line)
        if 'TORRENT_' in define and 'TORRENT_ABI_VERSION' not in define:
            print('\x1b[31mWARNING: possible ABI breakage in public struct! "%s" \x1b[34m %s:%d\x1b[0m' %
                  (define, filename, lno))
        elif define != '':
            print('\x1b[33msensitive define in public struct: "%s"\x1b[34m %s:%d\x1b[0m' % (define, filename, lno))

    if (line.startswith('#if') and (
            ' TORRENT_USE_ASSERTS' in line or
            ' TORRENT_USE_INVARIANT_CHECKS' in line or
            ' TORRENT_ASIO_DEBUGGING' in line) or
            line == '#if TORRENT_ABI_VERSION == 1' or
            line == '#if TORRENT_ABI_VERSION <= 2' or
            line == '#if TORRENT_ABI_VERSION < 3'):
        while lno < len(lines):
            line = lines[lno].strip()
            lno += 1
            if verbose:
                print('prep  %s' % line)
            if line.startswith('#endif'):
                end_if += 1
            if line.startswith('#if'):
                start_if += 1
            if line == '#else' and start_if - end_if == 1:
                break
            if start_if - end_if == 0:
                break
        return lno
    else:
        while line.endswith('\\') and lno < len(lines):
            line = lines[lno].strip()
            lno += 1
            if verbose:
                print('prep  %s' % line)

    return lno


for filename in files:
    h = open(filename)
    lines = h.read().split('\n')

    if verbose:
        print('\n=== %s ===\n' % filename)

    blanks = 0
    lno = 0
    orphaned_export = False

    while lno < len(lines):
        line = lines[lno].strip()

        if orphaned_export:
            print('ERROR: TORRENT_EXPORT without function or class!\n%s:%d\n%s' % (filename, lno, line))
            sys.exit(1)

        lno += 1

        if line == '':
            blanks += 1
            context = ''
            continue

        if 'TORRENT_EXPORT' in line.split() \
                and 'ifndef TORRENT_EXPORT' not in line \
                and 'define TORRENT_DEPRECATED_EXPORT TORRENT_EXPORT' not in line \
                and 'define TORRENT_EXPORT' not in line \
                and 'for TORRENT_EXPORT' not in line \
                and 'TORRENT_EXPORT TORRENT_CFG' not in line \
                and 'extern TORRENT_EXPORT ' not in line \
                and 'struct TORRENT_EXPORT ' not in line:
            orphaned_export = True
            if verbose:
                print('maybe orphaned: %s\n' % line)

        if line.startswith('//') and line[2:].strip() == 'OVERVIEW':
            # this is a section overview
            current_overview = ''
            while lno < len(lines):
                line = lines[lno].strip()
                lno += 1
                if not line.startswith('//'):
                    # end of overview
                    overviews[filename[11:]] = current_overview
                    current_overview = ''
                    break
                line = line[2:]
                if line.startswith(' '):
                    line = line[1:]
                current_overview += line + '\n'

        if line.startswith('//'):
            if verbose:
                print('desc  %s' % line)
            line = line[2:]
            if len(line) and line[0] == ' ':
                line = line[1:]
            context += line + '\n'
            continue

        if line.startswith('/*'):
            lno = consume_comment(lno - 1, lines)
            continue

        if line.startswith('#'):
            lno = consume_ifdef(lno - 1, lines)
            continue

        if (line == 'namespace aux {' or
                line == 'namespace ssl {' or
                line == 'namespace libtorrent { namespace aux {') \
                and not internal:
            lno = consume_block(lno - 1, lines)
            context = ''
            continue

        if 'namespace aux' in line and \
                '//' not in line.split('namespace')[0] and \
                '}' not in line.split('namespace')[1]:
            print('ERROR: whitespace preceding namespace declaration: %s:%d' % (filename, lno))
            sys.exit(1)

        if 'TORRENT_DEPRECATED' in line:
            if ('class ' in line or 'struct ' in line) and ';' not in line:
                lno = consume_block(lno - 1, lines)
                context = ''
            blanks += 1
            if verbose:
                print('xx    %s' % line)
            continue

        if looks_like_constant(line):
            if 'constexpr static' in line:
                print('ERROR: found "constexpr static", use "static constexpr" instead for consistency!\n%s:%d\n%s'
                      % (filename, lno, line))
                sys.exit(1)
            current_constant, lno = parse_constant(lno - 1, lines, filename)
            if current_constant is not None and is_visible(context):
                if 'TODO: ' in context:
                    print('TODO comment in public documentation: %s:%d' % (filename, lno))
                    sys.exit(1)
                current_constant['desc'] = context
                add_desc(context)
                if context == '':
                    print('WARNING: constant "%s" is not documented: \x1b[34m%s:%d\x1b[0m'
                          % (current_constant['name'], filename, lno))
                t = current_constant['type']
                if t in constants:
                    constants[t].append(current_constant)
                else:
                    constants[t] = [current_constant]
            continue

        if 'TORRENT_EXPORT ' in line or line.startswith('inline ') or line.startswith('template') or internal:
            if line.startswith('class ') or line.startswith('struct '):
                if not line.endswith(';'):
                    current_class, lno = parse_class(lno - 1, lines, filename)
                    if current_class is not None and is_visible(context):
                        if 'TODO: ' in context:
                            print('TODO comment in public documentation: %s:%d' % (filename, lno))
                            sys.exit(1)
                        current_class['desc'] = context
                        add_desc(context)
                        if context == '':
                            print('WARNING: class "%s" is not documented: \x1b[34m%s:%d\x1b[0m'
                                  % (current_class['name'], filename, lno))
                        classes.append(current_class)
                context = ''
                blanks += 1
                continue

            if looks_like_function(line):
                current_fun, lno = parse_function(lno - 1, lines, filename)
                if current_fun is not None and is_visible(context):
                    if context == '' and blanks == 0 and len(functions):
                        functions[-1]['signatures'].update(current_fun['signatures'])
                        functions[-1]['names'].update(current_fun['names'])
                    else:
                        if 'TODO: ' in context:
                            print('TODO comment in public documentation: %s:%d' % (filename, lno))
                            sys.exit(1)
                        current_fun['desc'] = context
                        add_desc(context)
                        if context == '':
                            print('WARNING: function "%s" is not documented: \x1b[34m%s:%d\x1b[0m'
                                  % (first_item(current_fun['names']), filename, lno))
                        functions.append(current_fun)
                    context = ''
                    blanks = 0
                continue

        if ('enum class ' not in line and 'class ' in line or 'struct ' in line) and ';' not in line:
            lno = consume_block(lno - 1, lines)
            context = ''
            blanks += 1
            continue

        if line.startswith('enum '):
            if not is_visible(context):
                consume_block(lno - 1, lines)
            else:
                current_enum, lno = parse_enum(lno - 1, lines, filename)
                if current_enum is not None and is_visible(context):
                    if 'TODO: ' in context:
                        print('TODO comment in public documentation: %s:%d' % (filename, lno))
                        sys.exit(1)
                    current_enum['desc'] = context
                    add_desc(context)
                    if context == '':
                        print('WARNING: enum "%s" is not documented: \x1b[34m%s:%d\x1b[0m'
                              % (current_enum['name'], filename, lno))
                    enums.append(current_enum)
            context = ''
            blanks += 1
            continue

        blanks += 1
        if verbose:
            if looks_like_forward_decl(line) \
                    or looks_like_blank(line) \
                    or looks_like_namespace(line):
                print('--    %s' % line)
            else:
                print('??    %s' % line)

        context = ''
    h.close()

# ====================================================================
#
#                               RENDER PART
#
# ====================================================================


def new_category(cat):
    return {'classes': [], 'functions': [], 'enums': [],
            'filename': 'reference-%s.rst' % cat.replace(' ', '_'),
            'constants': {}}


if dump:

    if verbose:
        print('\n===============================\n')

    for c in classes:
        print('\x1b[4m%s\x1b[0m %s\n{' % (c['type'], c['name']))
        for f in c['fun']:
            for s in f['signatures']:
                print('   %s' % s.replace('\n', '\n   '))

        if len(c['fun']) > 0 and len(c['fields']) > 0:
            print('')

        for f in c['fields']:
            for s in f['signatures']:
                print('   %s' % s)

        if len(c['fields']) > 0 and len(c['enums']) > 0:
            print('')

        for e in c['enums']:
            print('   \x1b[4menum\x1b[0m %s\n   {' % e['name'])
            for v in e['values']:
                print('      %s' % v['name'])
            print('   };')
        print('};\n')

    for f in functions:
        print('%s' % f['signature'])

    for e in enums:
        print('\x1b[4menum\x1b[0m %s\n{' % e['name'])
        for v in e['values']:
            print('   %s' % v['name'])
        print('};')

    for t, c in constants:
        print('\x1b[4mconstant\x1b[0m %s %s\n' % (e['type'], e['name']))

categories = {}

for c in classes:
    cat = categorize_symbol(c['name'], c['file'])
    if cat not in categories:
        categories[cat] = new_category(cat)

    if c['file'] in overviews:
        categories[cat]['overview'] = overviews[c['file']]

    filename = categories[cat]['filename'].replace('.rst', '.html') + '#'
    categories[cat]['classes'].append(c)
    symbols[c['name']] = filename + c['name']
    for f in c['fun']:
        for n in f['names']:
            symbols[n] = filename + n
            symbols[c['name'] + '::' + n] = filename + n

    for f in c['fields']:
        for n in f['names']:
            symbols[c['name'] + '::' + n] = filename + n

    for e in c['enums']:
        symbols[e['name']] = filename + e['name']
        symbols[c['name'] + '::' + e['name']] = filename + e['name']
        for v in e['values']:
            # symbols[v['name']] = filename + v['name']
            symbols[e['name'] + '::' + v['name']] = filename + v['name']
            symbols[c['name'] + '::' + v['name']] = filename + v['name']

for f in functions:
    cat = categorize_symbol(first_item(f['names']), f['file'])
    if cat not in categories:
        categories[cat] = new_category(cat)

    if f['file'] in overviews:
        categories[cat]['overview'] = overviews[f['file']]

    for n in f['names']:
        symbols[n] = categories[cat]['filename'].replace('.rst', '.html') + '#' + n
    categories[cat]['functions'].append(f)

for e in enums:
    cat = categorize_symbol(e['name'], e['file'])
    if cat not in categories:
        categories[cat] = new_category(cat)
    categories[cat]['enums'].append(e)
    filename = categories[cat]['filename'].replace('.rst', '.html') + '#'
    symbols[e['name']] = filename + e['name']
    for v in e['values']:
        symbols[e['name'] + '::' + v['name']] = filename + v['name']

for t, c in constants.items():
    for const in c:
        cat = categorize_symbol(t, const['file'])
        if cat not in categories:
            categories[cat] = new_category(cat)
        if t not in categories[cat]['constants']:
            categories[cat]['constants'][t] = [const]
        else:
            categories[cat]['constants'][t].append(const)
        filename = categories[cat]['filename'].replace('.rst', '.html') + '#'
        symbols[t + '::' + const['name']] = filename + t + '::' + const['name']
    symbols[t] = filename + t


def print_declared_in(out, o):
    out.write('Declared in "%s"\n\n' % print_link(o['file'], 'include/%s' % o['file']))
    print(dump_link_targets(), file=out)

# returns RST marked up string


def linkify_symbols(string):
    lines = string.split('\n')
    ret = []
    in_literal = False
    lno = 0
    return_string = ''
    for line in lines:
        lno += 1
        # don't touch headlines, i.e. lines whose
        # next line entirely contains one of =, - or .
        if (lno < len(lines) - 1):
            next_line = lines[lno]
        else:
            next_line = ''

        if '.. include:: ' in line:
            return_string += '\n'.join(ret)
            ret = [line]
            return_string += dump_link_targets() + '\n'
            continue

        if len(next_line) > 0 and lines[lno].replace('=', ''). \
                replace('-', '').replace('.', '') == '':
            ret.append(line)
            continue

        if line.startswith('|'):
            ret.append(line)
            continue
        if in_literal and not line.startswith('\t') and not line == '':
            # print('  end literal: "%s"' % line)
            in_literal = False
        if in_literal:
            # print('  literal: "%s"' % line)
            ret.append(line)
            continue
        if line.strip() == '.. parsed-literal::' or \
                line.strip().startswith('.. code::') or \
                (not line.strip().startswith('..') and line.endswith('::')):
            # print('  start literal: "%s"' % line)
            in_literal = True
        words = line.split(' ')

        for i in range(len(words)):
            # it's important to preserve leading
            # tabs, since that's relevant for
            # rst markup

            leading = ''
            w = words[i]

            if len(w) == 0:
                continue

            while len(w) > 0 and \
                    w[0] in ['\t', ' ', '(', '[', '{']:
                leading += w[0]
                w = w[1:]

            # preserve commas and dots at the end
            w = w.strip()
            trailing = ''

            if len(w) == 0:
                continue

            while len(w) > 1 and w[-1] in ['.', ',', ')'] and w[-2:] != '()':
                trailing = w[-1] + trailing
                w = w[:-1]

            link_name = w

            # print(w)

            if len(w) == 0:
                continue

            if link_name[-1] == '_':
                link_name = link_name[:-1]

            if w in symbols:
                link_name = link_name.replace('-', ' ')
                # print('  found %s -> %s' % (w, link_name))
                words[i] = leading + print_link(link_name, symbols[w]) + trailing
        ret.append(' '.join(words))
    return_string += '\n'.join(ret)
    return return_string


link_targets = []


def print_link(name, target):
    global link_targets
    link_targets.append(target)
    return "`%s`__" % name


def dump_link_targets(indent=''):
    global link_targets
    ret = '\n'
    for link in link_targets:
        ret += '%s__ %s\n' % (indent, link)
    link_targets = []
    return ret


def heading(string, c, indent=''):
    string = string.strip()
    return '\n' + indent + string + '\n' + indent + (c * len(string)) + '\n'


def render_enums(out, enums, print_declared_reference, header_level):
    for e in enums:
        print('.. raw:: html\n', file=out)
        print('\t<a name="%s"></a>' % e['name'], file=out)
        print('', file=out)
        dump_report_issue('enum ' + e['name'], out)
        print(heading('enum %s' % e['name'], header_level), file=out)

        print_declared_in(out, e)

        width = [len('name'), len('value'), len('description')]

        for i in range(len(e['values'])):
            e['values'][i]['desc'] = linkify_symbols(e['values'][i]['desc'])

        for v in e['values']:
            width[0] = max(width[0], len(v['name']))
            width[1] = max(width[1], len(v['val']))
            for d in v['desc'].split('\n'):
                width[2] = max(width[2], len(d))

        print('+-' + ('-' * width[0]) + '-+-' + ('-' * width[1]) + '-+-' + ('-' * width[2]) + '-+', file=out)
        print('| ' + 'name'.ljust(width[0]) + ' | ' + 'value'.ljust(width[1]) + ' | '
              + 'description'.ljust(width[2]) + ' |', file=out)
        print('+=' + ('=' * width[0]) + '=+=' + ('=' * width[1]) + '=+=' + ('=' * width[2]) + '=+', file=out)
        for v in e['values']:
            d = v['desc'].split('\n')
            if len(d) == 0:
                d = ['']
            print('| ' + v['name'].ljust(width[0]) + ' | ' + v['val'].ljust(width[1]) + ' | '
                  + d[0].ljust(width[2]) + ' |', file=out)
            for s in d[1:]:
                print('| ' + (' ' * width[0]) + ' | ' + (' ' * width[1]) + ' | ' + s.ljust(width[2]) + ' |', file=out)
            print('+-' + ('-' * width[0]) + '-+-' + ('-' * width[1]) + '-+-' + ('-' * width[2]) + '-+', file=out)
        print('', file=out)

        print(dump_link_targets(), file=out)


sections = \
    {
        'Core': 0,
        'DHT': 0,
        'Session': 0,
        'Torrent Handle': 0,
        'Torrent Info': 0,
        'Trackers': 0,
        'Settings': 0,
        'Torrent Status': 0,
        'Stats': 0,
        'Resume Data': 0,
        'Add Torrent': 0,

        'Bencoding': 1,
        'Bdecoding': 1,
        'Filter': 1,
        'Error Codes': 1,
        'Create Torrents': 1,

        'PeerClass': 2,
        'ed25519': 2,
        'Utility': 2,
        'Storage': 2,
        'Custom Storage': 2,
        'Plugins': 2,

        'Alerts': 3
    }


def print_toc(out, categories, s):

    main_toc = False

    for cat in categories:
        if (s != 2 and cat not in sections) or \
                (cat in sections and sections[cat] != s):
            continue

        if not main_toc:
            out.write('.. container:: main-toc\n\n')
            main_toc = True

        print('\t.. rubric:: %s\n' % cat, file=out)

        if 'overview' in categories[cat]:
            print('\t| overview__', file=out)

        for c in categories[cat]['classes']:
            print('\t| ' + print_link(c['name'], symbols[c['name']]), file=out)
        for f in categories[cat]['functions']:
            for n in f['names']:
                print('\t| ' + print_link(n, symbols[n]), file=out)
        for e in categories[cat]['enums']:
            print('\t| ' + print_link(e['name'], symbols[e['name']]), file=out)
        for t, c in categories[cat]['constants'].items():
            print('\t| ' + print_link(t, symbols[t]), file=out)
        print('', file=out)

        if 'overview' in categories[cat]:
            print('\t__ %s#overview' % categories[cat]['filename'].replace('.rst', '.html'), file=out)
        print(dump_link_targets('\t'), file=out)


def dump_report_issue(h, out):
    print(('.. raw:: html\n\n\t<span class="report-issue">[<a ' +
           'href="http://github.com/arvidn/libtorrent/issues/new?title=docs:{0}&labels=' +
           'documentation&body={1}">report issue</a>]</span>\n\n').format(
                urllib.parse.quote_plus(h),
                urllib.parse.quote_plus('Documentation under heading "' + h + '" could be improved')), file=out)


def render(out, category):

    classes = category['classes']
    functions = category['functions']
    enums = category['enums']
    constants = category['constants']

    if 'overview' in category:
        out.write('%s\n' % linkify_symbols(category['overview']))

    for c in classes:

        print('.. raw:: html\n', file=out)
        print('\t<a name="%s"></a>' % c['name'], file=out)
        print('', file=out)

        dump_report_issue('class ' + c['name'], out)
        out.write('%s\n' % heading(c['name'], '-'))
        print_declared_in(out, c)
        c['desc'] = linkify_symbols(c['desc'])
        out.write('%s\n' % c['desc'])
        print(dump_link_targets(), file=out)

        print('\n.. parsed-literal::\n\t', file=out)

        block = '\n%s\n{\n' % c['decl']
        for f in c['fun']:
            for s in f['signatures']:
                block += '   %s\n' % highlight_signature(s.replace('\n', '\n   '))

        if len(c['fun']) > 0 and len(c['enums']) > 0:
            block += '\n'

        first = True
        for e in c['enums']:
            if not first:
                block += '\n'
            first = False
            block += '   enum %s\n   {\n' % e['name']
            for v in e['values']:
                block += '      %s,\n' % v['name']
            block += '   };\n'

        if len(c['fun']) + len(c['enums']) > 0 and len(c['fields']):
            block += '\n'

        for f in c['fields']:
            for s in f['signatures']:
                block += '   %s\n' % highlight_name(s)

        block += '};'

        print(block.replace('\n', '\n\t') + '\n', file=out)

        for f in c['fun']:
            if f['desc'] == '':
                continue
            print('.. raw:: html\n', file=out)
            for n in f['names']:
                print('\t<a name="%s"></a>' % n, file=out)
            print('', file=out)
            h = ' '.join(f['names'])
            dump_report_issue('%s::[%s]' % (c['name'], h), out)
            print(heading(h, '.'), file=out)

            block = '.. parsed-literal::\n\n'

            for s in f['signatures']:
                block += highlight_signature(s.replace('\n', '\n   ')) + '\n'
            print('%s\n' % block.replace('\n', '\n\t'), file=out)
            f['desc'] = linkify_symbols(f['desc'])
            print('%s' % f['desc'], file=out)

            print(dump_link_targets(), file=out)

        render_enums(out, c['enums'], False, '.')

        for f in c['fields']:
            if f['desc'] == '':
                continue

            print('.. raw:: html\n', file=out)
            for n in f['names']:
                print('\t<a name="%s"></a>' % n, file=out)
            print('', file=out)
            h = ' '.join(f['names'])
            dump_report_issue('%s::[%s]' % (c['name'], h), out)
            print(h, file=out)
            f['desc'] = linkify_symbols(f['desc'])
            print('\t%s' % f['desc'].replace('\n', '\n\t'), file=out)

            print(dump_link_targets(), file=out)

    for f in functions:
        print('.. raw:: html\n', file=out)
        for n in f['names']:
            print('\t<a name="%s"></a>' % n, file=out)
        print('', file=out)
        h = ' '.join(f['names'])
        dump_report_issue(h, out)
        print(heading(h, '-'), file=out)
        print_declared_in(out, f)

        block = '.. parsed-literal::\n\n'
        for s in f['signatures']:
            block += highlight_signature(s) + '\n'

        print('%s\n' % block.replace('\n', '\n\t'), file=out)
        print(linkify_symbols(f['desc']), file=out)

        print(dump_link_targets(), file=out)

    render_enums(out, enums, True, '-')

    for t, c in constants.items():
        print('.. raw:: html\n', file=out)
        print('\t<a name="%s"></a>\n' % t, file=out)
        dump_report_issue(t, out)
        print(heading(t, '-'), file=out)
        print_declared_in(out, c[0])

        for v in c:
            print('.. raw:: html\n', file=out)
            print('\t<a name="%s::%s"></a>\n' % (t, v['name']), file=out)
            print(v['name'], file=out)
            v['desc'] = linkify_symbols(v['desc'])
            print('\t%s' % v['desc'].replace('\n', '\n\t'), file=out)
            print(dump_link_targets('\t'), file=out)

        print('', file=out)

    print(dump_link_targets(), file=out)

    for i in static_links:
        print(i, file=out)


if single_page_output:

    out = open('single-page-ref.rst', 'w+')
    out.write('''.. include:: header.rst

`home`__

__ reference.html

.. contents:: Table of contents
  :depth: 2
  :backlinks: none

''')

    for cat in categories:
        render(out, categories[cat])

    out.close()

else:

    out = open('reference.rst', 'w+')
    out.write('''=======================
reference documentation
=======================

''')

    out.write('`single-page version`__\n\n__ single-page-ref.html\n\n')

    for i in range(4):

        print_toc(out, categories, i)

    out.close()

    for cat in categories:
        out = open(categories[cat]['filename'], 'w+')

        out.write('''.. include:: header.rst

`home`__

__ reference.html

.. contents:: Table of contents
  :depth: 2
  :backlinks: none

''')

        render(out, categories[cat])

        out.close()

#       for s in symbols:
#           print(s)

    for i, o in list(preprocess_rst.items()):
        f = open(i, 'r')
        out = open(o, 'w+')
        print('processing %s -> %s' % (i, o))
        link = linkify_symbols(f.read())
        print(link, end=' ', file=out)

        print(dump_link_targets(), file=out)

        out.close()
        f.close()



# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4

from dataclasses import dataclass
import os
import platform
import subprocess
from time import monotonic


@dataclass(frozen=True)
class Metric:
    axis: str
    cumulative: bool


metrics = {
    "peak_nonpaged_pool": Metric("x1y1", False),
    "nonpaged_pool": Metric("x1y1", False),
    "num_page_faults": Metric("x1y2", True),
    "paged_pool": Metric("x1y1", False),
    "peak_paged_pool": Metric("x1y1", False),
    "peak_pagefile": Metric("x1y1", False),
    "peak_wset": Metric("x1y1", False),
    "private": Metric("x1y1", False),
    "rss": Metric("x1y1", False),
    "uss": Metric("x1y1", False),
    "data": Metric("x1y1", False),
    "shared": Metric("x1y1", False),
    "text": Metric("x1y1", False),
    "dirty": Metric("x1y1", False),
    "lib": Metric("x1y1", False),
    "vms": Metric("x1y1", False),
    "other_bytes": Metric("x1y1", True),
    "other_count": Metric("x1y2", True),
    "read_bytes": Metric("x1y1", True),
    "read_chars": Metric("x1y1", True),
    "read_count": Metric("x1y2", True),
    "write_bytes": Metric("x1y1", True),
    "write_chars": Metric("x1y1", True),
    "write_count": Metric("x1y2", True),
    "pfaults": Metric("x1y2", True),
    "pageins": Metric("x1y2", True),
    "minor_faults": Metric("x1y2", True),
    "major_faults": Metric("x1y2", True),
    "pss": Metric("x1y1", False),
    "pss_anon": Metric("x1y1", False),
    "pss_file": Metric("x1y1", False),
    "pss_shmem": Metric("x1y1", False),
    "shared_clean": Metric("x1y1", False),
    "shared_dirty": Metric("x1y1", False),
    "private_clean": Metric("x1y1", False),
    "private_dirty": Metric("x1y1", False),
    "referenced": Metric("x1y1", False),
    "anonymous": Metric("x1y1", False),
    "lazyfree": Metric("x1y1", False),
    "anonhugepages": Metric("x1y1", False),
    "shmempmdmapped": Metric("x1y1", False),
    "filepmdmapped": Metric("x1y1", False),
    "shared_hugetlb": Metric("x1y1", False),
    "private_hugetlb": Metric("x1y1", False),
    "swap": Metric("x1y1", False),
    "swappss": Metric("x1y1", False),
    "locked": Metric("x1y1", False),
}


@dataclass(frozen=True)
class Plot:
    name: str
    title: str
    ylabel: str
    y2label: str
    lines: list[str]


plots = [
    Plot(
        "memory",
        "libtorrent memory usage",
        "Memory Size",
        "",
        [
            "pss",
            "pss_file",
            "pss_anon",
            "rss",
            "dirty",
            "private_dirty",
            "private_clean",
            "lazyfree",
            "anonymous",
            "vms",
            "private",
            "paged_pool",
        ],
    ),
    Plot(
        "vm",
        "libtorrent vm stats",
        "",
        "count",
        [
            "pfaults",
            "pageins",
            "num_page_faults",
            "major_faults",
            "minor_faults",
        ],
    ),
    Plot(
        "read",
        "libtorrent disk I/O (read)",
        "Size",
        "count",
        [
            "read_bytes",
            "read_chars",
            "read_count",
        ],
    ),
    Plot(
        "write",
        "libtorrent disk I/O (write)",
        "Size",
        "count",
        [
            "write_bytes",
            "write_chars",
            "write_count",
        ],
    ),
]

if platform.system() == "Linux":

    def capture_sample(
        pid: int, start_time: float, output: dict[str, list[float]]
    ) -> None:
        try:
            with open(f"/proc/{pid}/smaps_rollup") as f:
                sample = f.read()
            with open(f"/proc/{pid}/stat") as f:
                sample2 = f.read()
            with open(f"/proc/{pid}/io") as f:
                sample3 = f.read()
            timestamp = monotonic() - start_time
        except Exception:
            return

        if "time" not in output:
            time_delta = 0.0
            output["time"] = [timestamp]
        else:
            time_delta = timestamp - output["time"][-1]
            output["time"].append(timestamp)

        for line in sample.split("\n"):
            if "[rollup]" in line:
                continue
            if line.strip() == "":
                continue

            key, value = line.split(":")
            val = int(value.split()[0].strip())
            key = key.strip().lower()

            if key not in output:
                output[key] = [val * 1024]
            else:
                output[key].append(val * 1024)

        stats = sample2.split()

        def add_counter(key: str, val: float) -> None:
            m = metrics[key]
            if key not in output:
                if m.cumulative:
                    output[key + "-raw"] = [val]
                    # we only have a single value, nothing to compare it against
                    val = 0
                output[key] = [val]
            else:
                if m.cumulative:
                    raw_val = val
                    val = (val - output[key + "-raw"][-1]) / time_delta
                    output[key + "-raw"].append(raw_val)
                output[key].append(val)

        add_counter("minor_faults", float(stats[9]))
        add_counter("major_faults", float(stats[11]))

        for line in sample3.split("\n"):
            if line == "":
                continue
            key, val2 = line.split(": ")
            if key == "rchar":
                add_counter("read_chars", float(val2))
            elif key == "wchar":
                add_counter("write_chars", float(val2))
            elif key == "read_bytes":
                add_counter("read_bytes", float(val2))
            elif key == "write_bytes":
                add_counter("write_bytes", float(val2))
            elif key == "syscr":
                add_counter("read_count", float(val2))
            elif key == "syscw":
                add_counter("write_count", float(val2))


# example output:
# 8affffff000-7fffba926000 ---p 00000000 00:00 0  [rollup]
#    Rss:               76932 kB
#    Pss:               17508 kB
#    Pss_Anon:          11376 kB
#    Pss_File:           6101 kB
#    Pss_Shmem:            30 kB
#    Shared_Clean:      65380 kB
#    Shared_Dirty:         88 kB
#    Private_Clean:        80 kB
#    Private_Dirty:     11384 kB
#    Referenced:        76932 kB
#    Anonymous:         11376 kB
#    LazyFree:              0 kB
#    AnonHugePages:         0 kB
#    ShmemPmdMapped:        0 kB
#    FilePmdMapped:         0 kB
#    Shared_Hugetlb:        0 kB
#    Private_Hugetlb:       0 kB
#    Swap:                  0 kB
#    SwapPss:               0 kB
#    Locked:                0 kB

else:
    import psutil

    def capture_sample(
        pid: int, start_time: float, output: dict[str, list[float]]
    ) -> None:
        try:
            p = psutil.Process(pid)
            mem = p.memory_full_info()
            io_cnt = p.io_counters()
            timestamp = monotonic() - start_time
        except Exception:
            return

        if "time" not in output:
            time_delta = 0.0
            output["time"] = [timestamp]
        else:
            time_delta = timestamp - output["time"][-1]
            output["time"].append(timestamp)

        for key in dir(mem):
            if key not in metrics:
                if not key.startswith("_") and key not in [
                    "pagefile",
                    "wset",
                    "count",
                    "index",
                ]:
                    print(f"missing key: {key}")
                continue

            val = getattr(mem, key)

            m = metrics[key]
            if key not in output:
                if m.cumulative:
                    output[key + "-raw"] = [val]
                    # we only have a single value, nothing to compare it against
                    val = 0
                output[key] = [val]
            else:
                if m.cumulative:
                    raw_val = val
                    val = (val - output[key + "-raw"][-1]) / time_delta
                    output[key + "-raw"].append(raw_val)

                output[key].append(val)

        for key in dir(io_cnt):
            if key not in metrics:
                if not key.startswith("_") and key not in [
                    "pagefile",
                    "wset",
                    "count",
                    "index",
                ]:
                    print(f"missing key: {key}")
                continue

            m = metrics[key]
            if key not in output:
                if m.cumulative:
                    output[key + "-raw"] = [val]
                    # we only have a single value, nothing to compare it against
                    val = 0
                output[key] = [val]
            else:
                if m.cumulative:
                    raw_val = val
                    val = (val - output[key + "-raw"][-1]) / time_delta
                    output[key + "-raw"].append(raw_val)

                output[key].append(val)


def print_output_to_file(out: dict[str, list[float]], filename: str) -> list[str]:
    if out == {}:
        return []

    with open(filename, "w+") as stats_output:
        non_zero_keys: set[str] = set()
        non_zero_keys.add("time")
        keys = out.keys()
        for key in keys:
            stats_output.write(f"{key} ")
        stats_output.write("\n")
        idx = 0
        while len(out["time"]) > idx:
            for key in keys:
                stats_output.write(f"{out[key][idx]:.2f} ")
                if out[key][idx] != 0:
                    non_zero_keys.add(key)
            stats_output.write("\n")
            idx += 1
    return [k if k in non_zero_keys else "" for k in keys]


def plot_output(filename: str, keys: list[str]) -> None:
    if "time" not in keys:
        return

    output_dir, in_file = os.path.split(filename)
    gnuplot_file = f"{output_dir}/plot_{in_file}.gnuplot"
    with open(gnuplot_file, "w+") as f:
        f.write("""set term png size 1200,700
set format y '%.0f'
set xlabel "time (s)"
set xrange [0:*]
set yrange [0:*]
set y2range [0:*]
set grid
""")

        for plot in plots:
            f.write(f"""set output "{in_file}-{plot.name}.png"
set title "{plot.title}"
set ylabel "{plot.ylabel} (MB)"
set y2label "{plot.y2label}"
{"set y2tics" if plot.y2label != "" else ""}
""")

            plot_string = "plot "
            tidx = keys.index("time") + 1
            idx = 0
            for p in keys:
                idx += 1
                if p == "time" or p == "":
                    continue

                if p not in plot.lines:
                    continue

                m = metrics[p]

                title = p.replace("_", "\\\\_")
                if m.cumulative:
                    title += "/s"

                divider = 1
                if m.axis == "x1y1":
                    divider = 1024 * 1024

                # escape underscores, since gnuplot interprets those as markup
                plot_string += (
                    f'"{in_file}" using {tidx}:(${idx}/{divider}) '
                    + f'title "{title}" axis {m.axis} with steps, \\\n'
                )
            if len(plot_string) > 5:
                plot_string = plot_string[0:-4] + "\n\n"
                f.write(plot_string)

    subprocess.check_output(["gnuplot", os.path.split(gnuplot_file)[1]], cwd=output_dir)



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4

from __future__ import print_function

import sys
import os

quantize = 100
max_rtt = 5000

f = open(sys.argv[1])
distribution = {}
num_messages = 0

for i in range(0, max_rtt, quantize):
    distribution[i] = 0

for line in f:
    time = int(line.split('\t')[1])
    if (time < 0 or time > max_rtt - quantize):
        continue
    num_messages += 1
    time //= quantize
    time *= quantize
    distribution[time] += 1

f = open('round_trip_distribution.log', 'w+')

for k, v in list(distribution.items()):
    print('%f %d' % ((k + (quantize / 2)) / 1000.0, v), file=f)
f.close()

f = open('round_trip_distribution.gnuplot', 'w+')

f.write('''
set term png size 1200,700
set title "Message round trip times"
set terminal postscript
set ylabel "# of requests"
set xlabel "Round trip time (seconds)"
set xrange [0:*]
set grid
set style fill solid border -1 pattern 2
set output "round_trip_distribution.ps"
set boxwidth %f
plot "round_trip_distribution.log" using 1:2 title "requests" with boxes

set terminal png small
set output "round_trip_distribution.png"
replot
''' % (float(quantize) / 1000.0))
f.close()

os.system('gnuplot round_trip_distribution.gnuplot')



#!/usr/bin/env python3
from __future__ import annotations

# Copyright Daniel Wallin 2006. Use, modification and distribution is
# subject to the Boost Software License, Version 1.0. (See accompanying
# file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

import atexit
import os.path
import sys
import time
from abc import ABC, abstractmethod
from collections.abc import Sequence, Mapping
from optparse import Values
from typing import Any, cast

import libtorrent as lt


class ConsoleABC(ABC):
    @abstractmethod
    def clear(self) -> None:
        pass

    @abstractmethod
    def write(self, line: str) -> None:
        pass

    @abstractmethod
    def sleep_and_input(self, seconds: float) -> bytes | str | None:
        pass


class WindowsConsole(ConsoleABC):
    def __init__(self) -> None:
        import Console

        self.console = Console.getconsole()

    def clear(self) -> None:
        self.console.page()

    def write(self, line: str) -> None:
        self.console.write(line)

    def sleep_and_input(self, seconds: float) -> bytes | None:
        import msvcrt

        time.sleep(seconds)
        if msvcrt.kbhit():
            return msvcrt.getch()
        return None


class UnixConsole(ConsoleABC):
    def __init__(self) -> None:
        import termios

        self.fd = sys.stdin
        self.old = termios.tcgetattr(self.fd.fileno())
        new = termios.tcgetattr(self.fd.fileno())
        new[3] = new[3] & ~termios.ICANON
        new[6][termios.VTIME] = 0
        new[6][termios.VMIN] = 1
        termios.tcsetattr(self.fd.fileno(), termios.TCSADRAIN, new)

        atexit.register(self._onexit)

    def _onexit(self) -> None:
        import termios

        termios.tcsetattr(self.fd.fileno(), termios.TCSADRAIN, self.old)

    def clear(self) -> None:
        sys.stdout.write('\033[2J\033[0;0H')
        sys.stdout.flush()

    def write(self, line: str) -> None:
        sys.stdout.write(line)
        sys.stdout.flush()

    def sleep_and_input(self, seconds: float) -> str | None:
        import select

        read, __, __ = select.select(
            [self.fd.fileno()], [], [], seconds)
        if len(read) > 0:
            return self.fd.read(1)
        return None


def write_line(console: ConsoleABC, line: str) -> None:
    console.write(line)


def add_suffix(val: float) -> str:
    prefix = ['B', 'kB', 'MB', 'GB', 'TB']
    for i in range(len(prefix)):
        if abs(val) < 1000:
            if i == 0:
                return '%5.3g%s' % (val, prefix[i])
            else:
                return '%4.3g%s' % (val, prefix[i])
        val /= 1000

    return '%6.3gPB' % val


def progress_bar(progress: float, width: int) -> str:
    assert(progress <= 1)
    progress_chars = int(progress * width + 0.5)
    return progress_chars * '#' + (width - progress_chars) * '-'


def print_peer_info(console: ConsoleABC, peers: list[lt.peer_info]) -> None:

    out = (' down    (total )   up     (total )'
           '  q  r flags  block progress  client\n')

    for p in peers:

        out += '%s/s ' % add_suffix(p.down_speed)
        out += '(%s) ' % add_suffix(p.total_download)
        out += '%s/s ' % add_suffix(p.up_speed)
        out += '(%s) ' % add_suffix(p.total_upload)
        out += '%2d ' % p.download_queue_length
        out += '%2d ' % p.upload_queue_length

        out += 'I' if p.flags & lt.peer_info.interesting else '.'
        out += 'C' if p.flags & lt.peer_info.choked else '.'
        out += 'i' if p.flags & lt.peer_info.remote_interested else '.'
        out += 'c' if p.flags & lt.peer_info.remote_choked else '.'
        out += 'e' if p.flags & lt.peer_info.supports_extensions else '.'
        out += 'l' if p.flags & lt.peer_info.local_connection else 'r'
        out += ' '

        if p.downloading_piece_index >= 0:
            assert(p.downloading_progress <= p.downloading_total)
            out += progress_bar(float(p.downloading_progress) /
                                p.downloading_total, 15)
        else:
            out += progress_bar(0, 15)
        out += ' '

        if p.flags & lt.peer_info.handshake:
            id = 'waiting for handshake'
        elif p.flags & lt.peer_info.connecting:
            id = 'connecting to peer'
        else:
            try:
                id = p.client.decode()
            except ValueError:
                id = p.client.hex()

        out += '%s\n' % id[:10]

    write_line(console, out)


def print_download_queue(console: ConsoleABC, download_queue: Sequence[Mapping[str, Any]]) -> None:
    out = ""

    for e in download_queue:
        out += '%4d: [' % e['piece_index']
        for b in e['blocks']:
            s = b['state']
            if s == 3:
                out += '#'
            elif s == 2:
                out += '='
            elif s == 1:
                out += '-'
            else:
                out += ' '
        out += ']\n'

    write_line(console, out)


def add_torrent(ses: lt.session, filename: str, options: Values) -> None:
    atp = lt.add_torrent_params()
    save_path = cast(str, options.save_path)
    if filename.startswith('magnet:'):
        atp = lt.parse_magnet_uri(filename)
    else:
        ti = lt.torrent_info(filename)
        resume_file = os.path.join(save_path, ti.name() + '.fastresume')
        try:
            atp = lt.read_resume_data(open(resume_file, 'rb').read())
        except Exception as e:
            print('failed to open resume file "%s": %s' % (resume_file, e))
        atp.ti = ti

    atp.save_path = save_path
    atp.storage_mode = lt.storage_mode_t.storage_mode_sparse
    atp.flags |= lt.torrent_flags.duplicate_is_error \
        | lt.torrent_flags.auto_managed \
        | lt.torrent_flags.duplicate_is_error
    ses.async_add_torrent(atp)


def main() -> None:
    from optparse import OptionParser

    parser = OptionParser()

    parser.add_option('-p', '--port', type='int', help='set listening port')

    parser.add_option(
        '-i', '--listen-interface', type='string',
        help='set interface for incoming connections', )

    parser.add_option(
        '-o', '--outgoing-interface', type='string',
        help='set interface for outgoing connections')

    parser.add_option(
        '-d', '--max-download-rate', type='float',
        help='the maximum download rate given in kB/s. 0 means infinite.')

    parser.add_option(
        '-u', '--max-upload-rate', type='float',
        help='the maximum upload rate given in kB/s. 0 means infinite.')

    parser.add_option(
        '-s', '--save-path', type='string',
        help='the path where the downloaded file/folder should be placed.')

    parser.add_option(
        '-r', '--proxy-host', type='string',
        help='sets HTTP proxy host and port (separated by \':\')')

    parser.set_defaults(
        port=6881,
        listen_interface='0.0.0.0',
        outgoing_interface='',
        max_download_rate=0,
        max_upload_rate=0,
        save_path='.',
        proxy_host=''
    )

    (options, args) = parser.parse_args()

    if options.port < 0 or options.port > 65525:
        options.port = 6881

    options.max_upload_rate *= 1000
    options.max_download_rate *= 1000

    if options.max_upload_rate <= 0:
        options.max_upload_rate = -1
    if options.max_download_rate <= 0:
        options.max_download_rate = -1

    params = lt.session_params()
    params.settings.update({
        "user_agent": 'python_client/' + lt.__version__,
        "listen_interfaces": '%s:%d' % (options.listen_interface, options.port),
        "download_rate_limit": int(options.max_download_rate),
        "upload_rate_limit": int(options.max_upload_rate),
        "alert_mask": lt.alert.category_t.all_categories,
        "outgoing_interfaces": options.outgoing_interface,
    })

    if options.proxy_host != '':
        params.settings["proxy_hostname"] = options.proxy_host.split(':')[0]
        params.settings["proxy_type"] = lt.proxy_type_t.http
        params.settings["proxy_port"] = options.proxy_host.split(':')[1]

    ses = lt.session(params)

    # map torrent_handle to torrent_status
    torrents: dict[lt.torrent_handle, lt.torrent_status] = {}
    alerts_log = []

    for f in args:
        add_torrent(ses, f, options)

    console: ConsoleABC
    if os.name == 'nt':
        console = WindowsConsole()
    else:
        console = UnixConsole()

    alive = True
    while alive:
        console.clear()

        out = ''

        for h, t in torrents.items():
            out += 'name: %-40s\n' % t.name[:40]

            if t.state != lt.torrent_status.seeding:
                state_str = ['queued', 'checking', 'downloading metadata',
                             'downloading', 'finished', 'seeding',
                             '', 'checking fastresume']
                out += state_str[t.state] + ' '

                out += '%5.4f%% ' % (t.progress * 100)
                out += progress_bar(t.progress, 49)
                out += '\n'

                out += 'total downloaded: %d Bytes\n' % t.total_done
                out += 'peers: %d seeds: %d distributed copies: %d\n' % \
                    (t.num_peers, t.num_seeds, t.distributed_copies)
                out += '\n'

            out += 'download: %s/s (%s) ' \
                % (add_suffix(t.download_rate), add_suffix(t.total_download))
            out += 'upload: %s/s (%s) ' \
                % (add_suffix(t.upload_rate), add_suffix(t.total_upload))

            if t.state != lt.torrent_status.seeding:
                out += 'info-hash: %s\n' % t.info_hashes
                out += 'next announce: %s\n' % t.next_announce
                out += 'tracker: %s\n' % t.current_tracker

            write_line(console, out)

            print_peer_info(console, t.handle.get_peer_info())
            print_download_queue(console, t.handle.get_download_queue())

            if t.state != lt.torrent_status.seeding:
                try:
                    out = '\n'
                    fp = h.file_progress()
                    ti = cast(lt.torrent_info, t.torrent_file)
                    for idx, p in enumerate(fp):
                        out += progress_bar(p / float(ti.files().file_size(idx)), 20)
                        out += ' ' + ti.files().file_path(idx) + '\n'
                    write_line(console, out)
                except Exception:
                    pass

        write_line(console, 76 * '-' + '\n')
        write_line(console, '(q)uit), (p)ause), (u)npause), (r)eannounce\n')
        write_line(console, 76 * '-' + '\n')

        alerts = ses.pop_alerts()
        for a in alerts:
            alerts_log.append(a.message())

            # add new torrents to our list of torrent_status
            if isinstance(a, lt.add_torrent_alert):
                h = a.handle
                h.set_max_connections(60)
                h.set_max_uploads(-1)
                torrents[h] = h.status()

            # update our torrent_status array for torrents that have
            # changed some of their state
            if isinstance(a, lt.state_update_alert):
                for s in a.status:
                    torrents[s.handle] = s

        if len(alerts_log) > 20:
            alerts_log = alerts_log[-20:]

        for a_log in alerts_log:
            write_line(console, a_log + '\n')

        c = console.sleep_and_input(0.5)

        ses.post_torrent_updates()
        if not c:
            continue

        if c == 'r':
            for h in torrents:
                h.force_reannounce()
        elif c == 'q':
            alive = False
        elif c == 'p':
            for h in torrents:
                h.pause()
        elif c == 'u':
            for h in torrents:
                h.resume()

    ses.pause()
    for h, t in torrents.items():
        if not h.is_valid() or not t.has_metadata:
            continue
        h.save_resume_data()

    while len(torrents) > 0:
        alerts = ses.pop_alerts()
        for a in alerts:
            if isinstance(a, lt.save_resume_data_alert):
                print(a)
                data = lt.write_resume_data_buf(a.params)
                h = a.handle
                if h in torrents:
                    open(os.path.join(options.save_path, torrents[h].name + '.fastresume'), 'wb').write(data)
                    del torrents[h]

            if isinstance(a, lt.save_resume_data_failed_alert):
                h = a.handle
                if h in torrents:
                    print('failed to save resume data for ', torrents[h].name)
                    del torrents[h]
        time.sleep(0.5)


main()



#!/usr/bin/env python3

"""Minimal non-feature complete socks proxy"""

import socket
from struct import pack, unpack
import threading
import sys
import traceback

from socketserver import StreamRequestHandler, ThreadingTCPServer


def debug(s):
    print('socks.py: ', s)
    sys.stdout.flush()


def error(s):
    print('socks.py, ERROR: ', s)
    sys.stdout.flush()


class MyTCPServer(ThreadingTCPServer):
    allow_reuse_address = True

    def handle_timeout(self):
        raise Exception('socks.py: timeout')


CLOSE = object()

VERSION = b'\x05'
NOAUTH = b'\x00'
USERPASS = b'\x02'
CONNECT = b'\x01'
UDP_ASSOCIATE = b'\x03'
IPV4 = b'\x01'
IPV6 = b'\x04'
DOMAIN_NAME = b'\x03'
SUCCESS = b'\x00'

password = None
username = None
allow_v4 = False


def send(dest, msg):
    if msg == CLOSE:
        try:
            dest.shutdown(socket.SHUT_WR)
        except Exception:
            pass
        dest.close()
        return 0
    else:
        return dest.sendall(msg)


def recv(source, n):
    data = source.recv(n)
    if data == b'':
        return CLOSE
    else:
        return data


def forward(source, dest, name):
    while True:
        data = recv(source, 4000)
        if data == CLOSE:
            send(dest, CLOSE)
            debug('%s hung up' % name)
            return
#        debug('Forwarding (%d) %r' % (len(data), data))
        send(dest, data)


def spawn_forwarder(source, dest, name):
    t = threading.Thread(target=forward, args=(source, dest, name))
    t.daemon = True
    t.start()


class SocksHandler(StreamRequestHandler):
    """Highly feature incomplete SOCKS 5 implementation"""

    def close_request(self):
        self.server.close_request(self.request)

    def read(self, n):
        data = b''
        while len(data) < n:
            extra = self.rfile.read(n)
            if extra == b'':
                raise Exception('Connection closed')
            data += extra
        return data

    def handle(self):
        try:
            self.inner_handle()
        except Exception:
            error("Unhandled exception")
            traceback.print_exc(file=sys.stdout)
            sys.stdout.flush()

    def inner_handle(self):
        # IMPROVEMENT: Report who requests are from in logging
        # IMPROVEMENT: Timeout on client
        debug('Connection - authenticating')
        version = self.read(1)

        if allow_v4 and version == b'\x04':
            cmd = self.read(1)
            if cmd != CONNECT and cmd != UDP_ASSOCIATE:
                error('Only supports connect and udp-associate method not (%r) closing' % cmd)
                self.close_request()
                return

            raw_dest_port = self.read(2)
            dest_port, = unpack('>H', raw_dest_port)

            raw_dest_address = self.read(4)
            dest_address = '.'.join(map(str, unpack('>4B', raw_dest_address)))

            user_id = b''
            c = self.read(1)
            while c != b'\0':
                user_id += c
                c = self.read(1)

            outbound_sock = socket.socket(socket.AF_INET)
            out_address = socket.getaddrinfo(dest_address, dest_port)[0][4]
            debug("Creating forwarder connection to %s:%d" % (out_address[0], out_address[1]))
            outbound_sock.connect(out_address)

            self.send_reply_v4(outbound_sock.getsockname())

            spawn_forwarder(outbound_sock, self.request, 'destination')
            forward(self.request, outbound_sock, 'client')
            return

        if version != b'\x05':
            error('Wrong version number (%r) closing...' % version)
            self.close_request()
            return

        nmethods = ord(self.read(1))
        method_list = self.read(nmethods)

        global password
        global username

        if password is None and NOAUTH in method_list:
            self.send_no_auth_method()
            debug('Authenticated (no-auth)')
        elif USERPASS in method_list:
            self.send_user_pass_auth_method()
            auth_version = self.read(1)
            if auth_version != b'\x01':
                error('Wrong sub-negotiation version number (%r) closing...' % version)
                self.close_request()
                return
            usr_len = ord(self.read(1))
            usr_name = self.read(usr_len)
            pwd_len = ord(self.read(1))
            pwd = self.read(pwd_len)

            if usr_name != username.encode() or pwd != password.encode():
                error('Invalid username or password')
                self.close_request()
                return
            debug('Authenticated (user/password)')
            self.send_authenticated()
        else:
            error('Server only supports NOAUTH and user/pass')
            self.send_no_method()
            return

        # If we were authenticating it would go here
        version = self.read(1)
        cmd = self.read(1)
        zero = self.read(1)
        address_type = self.read(1)
        if version != b'\x05':
            error('Wrong version number (%r) closing...' % version)
            self.close_request()
        elif cmd != CONNECT and cmd != UDP_ASSOCIATE:
            error('Only supports connect and udp-associate method not (%r) closing' % cmd)
            self.close_request()
        elif zero != b'\x00':
            error('Mangled request. Reserved field (%r) is not null' % zero)
            self.close_request()

        if address_type == IPV4:
            raw_dest_address = self.read(4)
            dest_address = '.'.join(map(str, unpack('>4B', raw_dest_address)))
        elif address_type == IPV6:
            raw_dest_address = self.read(16)
            dest_address = ":".join([hex(x)[2:] for x in unpack('>8H', raw_dest_address)])
        elif address_type == DOMAIN_NAME:
            dns_length = ord(self.read(1))
            dns_name = self.read(dns_length)
            dest_address = dns_name
        else:
            error('Unknown addressing (%r)' % address_type)
            self.close_request()

        raw_dest_port = self.read(2)
        dest_port, = unpack('>H', raw_dest_port)

        if address_type == IPV6:
            outbound_sock = socket.socket(socket.AF_INET6)
        else:
            outbound_sock = socket.socket(socket.AF_INET)
        try:
            out_address = socket.getaddrinfo(dest_address, dest_port)[0][4]
        except Exception as e:
            error('%s' % e)
            traceback.print_exc(file=sys.stdout)
            return

        if cmd == UDP_ASSOCIATE:
            debug("no UDP support yet, closing")
            return

        debug("Creating forwarder connection to %s:%d" % (out_address[0], out_address[1]))

        try:
            outbound_sock.connect(out_address)
        except Exception as e:
            error('%s' % e)
            traceback.print_exc(file=sys.stdout)
            return

        if address_type == IPV6:
            self.send_reply6(outbound_sock.getsockname())
        else:
            self.send_reply(outbound_sock.getsockname())

        spawn_forwarder(outbound_sock, self.request, 'destination')
        try:
            forward(self.request, outbound_sock, 'client')
        except Exception as e:
            error('%s' % e)
            traceback.print_exc(file=sys.stdout)

    def send_reply_v4(self, xxx_todo_changeme):
        (bind_addr, bind_port) = xxx_todo_changeme
        self.wfile.write(b'\0\x5a\0\0\0\0\0\0')
        self.wfile.flush()

    def send_reply(self, xxx_todo_changeme1):
        (bind_addr, bind_port) = xxx_todo_changeme1
        bind_tuple = tuple(map(int, bind_addr.split('.')))
        full_address = bind_tuple + (bind_port,)
        debug('Setting up forwarding port %r' % (full_address,))
        msg = pack('>cccc4BH', VERSION, SUCCESS, b'\x00', IPV4, *full_address)
        self.wfile.write(msg)

    def send_reply6(self, xxx_todo_changeme2):
        (bind_addr, bind_port, unused1, unused2) = xxx_todo_changeme2
        bind_tuple = tuple([int(x, 16) for x in bind_addr.split(':')])
        full_address = bind_tuple + (bind_port,)
        debug('Setting up forwarding port %r' % (full_address,))
        msg = pack('>cccc8HH', VERSION, SUCCESS, b'\x00', IPV6, *full_address)
        self.wfile.write(msg)

    def send_no_method(self):
        self.wfile.write(b'\x05\xff')
        self.close_request()

    def send_no_auth_method(self):
        self.wfile.write(b'\x05\x00')
        self.wfile.flush()

    def send_user_pass_auth_method(self):
        self.wfile.write(b'\x05\x02')
        self.wfile.flush()

    def send_authenticated(self):
        self.wfile.write(b'\x01\x00')
        self.wfile.flush()


if __name__ == '__main__':

    debug('starting socks.py %s' % " ".join(sys.argv))
    debug('python version: %s' % sys.version_info.__str__())
    listen_port = 8002
    i = 1
    while i < len(sys.argv):
        if sys.argv[i] == '--username':
            username = sys.argv[i + 1]
            i += 1
        elif sys.argv[i] == '--password':
            password = sys.argv[i + 1]
            i += 1
        elif sys.argv[i] == '--port':
            listen_port = int(sys.argv[i + 1])
            i += 1
        elif sys.argv[i] == '--allow-v4':
            allow_v4 = True
        else:
            if sys.argv[i] != '--help':
                debug('unknown option "%s"' % sys.argv[i])
            print('usage: socks.py [--username <user> --password <password>] [--port <listen-port>]')
            sys.stdout.flush()
            sys.exit(1)
        i += 1

    debug('Listening on port %d...' % listen_port)
    server = MyTCPServer(('localhost', listen_port), SocksHandler)
    server.timeout = 190
    while True:
        server.handle_request()



import libtorrent as lt

import hashlib
import random

PIECE_LENGTH = 16384
NAME = b"test.txt"
LEN = PIECE_LENGTH * 9 + 1000
# Use 7-bit data so we can test piece data as either bytes or str
DATA = bytes(random.getrandbits(7) for _ in range(LEN))
PIECES = [DATA[i:i + PIECE_LENGTH] for i in range(0, LEN, PIECE_LENGTH)]

INFO_DICT = {
        b"name": NAME,
        b"piece length": PIECE_LENGTH,
        b"length": len(DATA),
        b"pieces": b"".join(hashlib.sha1(p).digest() for p in PIECES),
    }

DICT = {
    b"info": INFO_DICT,
}


def get_infohash_bytes() -> bytes:
    return hashlib.sha1(lt.bencode(INFO_DICT)).digest()


def get_infohash() -> str:
    return get_infohash_bytes().hex()


def get_sha1_hash() -> lt.sha1_hash:
    return lt.sha1_hash(get_infohash_bytes())
